# coding: utf-8

import os.path
from textwrap import dedent
try:
    import unittest2 as unittest
except ImportError:
    import unittest

from rope.base import exceptions
from rope.contrib.codeassist import (get_definition_location, get_doc,
                                     starting_expression, code_assist,
                                     sorted_proposals, starting_offset,
                                     get_calltip, get_canonical_path)
from ropetest import testutils

try:
    unicode
except NameError:
    unicode = str


class CodeAssistTest(unittest.TestCase):

    def setUp(self):
        super(CodeAssistTest, self).setUp()
        self.project = testutils.sample_project()

    def tearDown(self):
        testutils.remove_project(self.project)
        super(CodeAssistTest, self).tearDown()

    def _assist(self, code, offset=None, **args):
        if offset is None:
            offset = len(code)
        return code_assist(self.project, code, offset, **args)

    def test_simple_assist(self):
        self._assist('', 0)

    def assert_completion_in_result(self, name, scope, result, type=None):
        for proposal in result:
            if proposal.name == name:
                self.assertEqual(scope, proposal.scope,
                                 "proposal <%s> has wrong scope, expected "
                                 "%r, got %r" % (name, scope, proposal.scope))
                if type is not None:
                    self.assertEqual(type, proposal.type,
                                     "proposal <%s> has wrong type, expected "
                                     "%r, got %r" %
                                     (name, type, proposal.type))
                return
        self.fail('completion <%s> not proposed' % name)

    def assert_completion_not_in_result(self, name, scope, result):
        for proposal in result:
            if proposal.name == name and proposal.scope == scope:
                self.fail('completion <%s> was proposed' % name)

    def test_completing_global_variables(self):
        code = 'my_global = 10\nt = my'
        result = self._assist(code)
        self.assert_completion_in_result('my_global', 'global', result)

    def test_not_proposing_unmatched_vars(self):
        code = 'my_global = 10\nt = you'
        result = self._assist(code)
        self.assert_completion_not_in_result('my_global', 'global', result)

    def test_not_proposing_unmatched_vars_with_underlined_starting(self):
        code = 'my_global = 10\nt = your_'
        result = self._assist(code)
        self.assert_completion_not_in_result('my_global', 'global', result)

    def test_not_proposing_local_assigns_as_global_completions(self):
        code = 'def f():    my_global = 10\nt = my_'
        result = self._assist(code)
        self.assert_completion_not_in_result('my_global', 'global', result)

    def test_proposing_functions(self):
        code = 'def my_func():    return 2\nt = my_'
        result = self._assist(code)
        self.assert_completion_in_result('my_func', 'global', result)

    def test_proposing_classes(self):
        code = 'class Sample(object):    pass\nt = Sam'
        result = self._assist(code)
        self.assert_completion_in_result('Sample', 'global', result)

    def test_proposing_each_name_at_most_once(self):
        code = 'variable = 10\nvariable = 20\nt = vari'
        result = self._assist(code)
        count = len([x for x in result
                     if x.name == 'variable' and x.scope == 'global'])
        self.assertEqual(1, count)

    def test_throwing_exception_in_case_of_syntax_errors(self):
        code = 'sample (sdf+)\n'
        with self.assertRaises(exceptions.ModuleSyntaxError):
            self._assist(code, maxfixes=0)

    def test_fixing_errors_with_maxfixes(self):
        code = 'def f():\n    sldj sldj\ndef g():\n    ran'
        result = self._assist(code, maxfixes=2)
        self.assertTrue(len(result) > 0)

    def test_ignoring_errors_in_current_line(self):
        code = 'def my_func():\n    return 2\nt = '
        result = self._assist(code)
        self.assert_completion_in_result('my_func', 'global', result)

    def test_not_reporting_variables_in_current_line(self):
        code = 'def my_func():    return 2\nt = my_'
        result = self._assist(code)
        self.assert_completion_not_in_result('my_', 'global', result)

    def test_completion_result(self):
        code = 'my_global = 10\nt = my'
        self.assertEqual(len(code) - 2, starting_offset(code, len(code)))

    def test_completing_imported_names(self):
        code = 'import sys\na = sy'
        result = self._assist(code)
        self.assert_completion_in_result('sys', 'imported', result)

    def test_completing_imported_names_with_as(self):
        code = 'import sys as mysys\na = mys'
        result = self._assist(code)
        self.assert_completion_in_result('mysys', 'imported', result)

    def test_not_completing_imported_names_with_as(self):
        code = 'import sys as mysys\na = sy'
        result = self._assist(code)
        self.assert_completion_not_in_result('sys', 'global', result)

    def test_including_matching_builtins_types(self):
        code = 'my_var = Excep'
        result = self._assist(code)
        self.assert_completion_in_result('Exception', 'builtin', result)
        self.assert_completion_not_in_result('zip', 'builtin', result)

    def test_including_matching_builtins_functions(self):
        code = 'my_var = zi'
        result = self._assist(code)
        self.assert_completion_in_result('zip', 'builtin', result)

    def test_builtin_instances(self):
        # ``import_dynload_stdmods`` pref is disabled for test project.
        # we need to have it enabled to make pycore._find_module()
        # load ``sys`` module.
        self.project.prefs['import_dynload_stdmods'] = True
        code = 'from sys import stdout\nstdout.wr'
        result = self._assist(code)
        self.assert_completion_in_result('write', 'builtin', result)
        self.assert_completion_in_result('writelines', 'builtin', result)

    def test_including_keywords(self):
        code = 'fo'
        result = self._assist(code)
        self.assert_completion_in_result('for', 'keyword', result)

    def test_not_reporting_proposals_after_dot(self):
        code = 'a_dict = {}\nkey = 3\na_dict.ke'
        result = self._assist(code)
        self.assert_completion_not_in_result('key', 'global', result)

    def test_proposing_local_variables_in_functions(self):
        code = 'def f(self):\n    my_var = 10\n    my_'
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'local', result)

    def test_local_variables_override_global_ones(self):
        code = 'my_var = 20\ndef f(self):\n    my_var = 10\n    my_'
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'local', result)

    def test_not_including_class_body_variables(self):
        code = 'class C(object):\n    my_var = 20\n' \
               '    def f(self):\n        a = 20\n        my_'
        result = self._assist(code)
        self.assert_completion_not_in_result('my_var', 'local', result)

    def test_nested_functions(self):
        code = 'def my_func():\n    func_var = 20\n    ' \
               'def inner_func():\n        a = 20\n        func'
        result = self._assist(code)
        self.assert_completion_in_result('func_var', 'local', result)

    def test_scope_endpoint_selection(self):
        code = "def my_func():\n    func_var = 20\n"
        result = self._assist(code)
        self.assert_completion_not_in_result('func_var', 'local', result)

    def test_scope_better_endpoint_selection(self):
        code = "if True:\n    def f():\n        my_var = 10\n    my_"
        result = self._assist(code)
        self.assert_completion_not_in_result('my_var', 'local', result)

    def test_imports_inside_function(self):
        code = "def f():\n    import sys\n    sy"
        result = self._assist(code)
        self.assert_completion_in_result('sys', 'imported', result)

    def test_imports_inside_function_dont_mix_with_globals(self):
        code = "def f():\n    import sys\nsy"
        result = self._assist(code)
        self.assert_completion_not_in_result('sys', 'local', result)

    def test_nested_classes_local_names(self):
        code = 'global_var = 10\n' \
               'def my_func():\n' \
               '    func_var = 20\n' \
               '    class C(object):\n' \
               '        def another_func(self):\n' \
               '            local_var = 10\n' \
               '            func'
        result = self._assist(code)
        self.assert_completion_in_result('func_var', 'local', result)

    def test_nested_classes_global(self):
        code = 'global_var = 10\n' \
               'def my_func():\n' \
               '    func_var = 20\n' \
               '    class C(object):\n' \
               '        def another_func(self):\n' \
               '            local_var = 10\n' \
               '            globa'
        result = self._assist(code)
        self.assert_completion_in_result('global_var', 'global', result)

    def test_nested_classes_global_function(self):
        code = 'global_var = 10\n' \
               'def my_func():\n' \
               '    func_var = 20\n' \
               '    class C(object):\n' \
               '        def another_func(self):\n' \
               '            local_var = 10\n' \
               '            my_f'
        result = self._assist(code)
        self.assert_completion_in_result('my_func', 'global', result)

    def test_proposing_function_parameters_in_functions(self):
        code = 'def my_func(my_param):\n    my_var = 20\n    my_'
        result = self._assist(code)
        self.assert_completion_in_result('my_param', 'local', result)

    def test_proposing_function_keyword_parameters_in_functions(self):
        code = 'def my_func(my_param, *my_list, **my_kws):\n' \
               '    my_var = 20\n' \
               '    my_'
        result = self._assist(code)
        self.assert_completion_in_result('my_param', 'local', result)
        self.assert_completion_in_result('my_list', 'local', result)
        self.assert_completion_in_result('my_kws', 'local', result)

    def test_not_proposing_unmatching_function_parameters_in_functions(self):
        code = "def my_func(my_param):\n    my_var = 20\n    you_"
        result = self._assist(code)
        self.assert_completion_not_in_result('my_param', 'local', result)

    def test_ignoring_current_statement(self):
        code = "my_var = 10\nmy_tuple = (10, \n           my_"
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'global', result)

    def test_ignoring_current_statement_brackets_continuation(self):
        code = "my_var = 10\n'hello'[10:\n        my_"
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'global', result)

    def test_ignoring_current_statement_explicit_continuation(self):
        code = "my_var = 10\nmy_var2 = 2 + \\\n          my_"
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'global', result)

    def test_ignor_current_statement_while_the_first_stmnt_of_the_block(self):
        code = "my_var = 10\ndef f():\n    my_"
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'global', result)

    def test_ignor_current_stmnt_while_current_line_ends_with_a_colon(self):
        code = "my_var = 10\nif my_:\n    pass"
        result = self._assist(code, 18)
        self.assert_completion_in_result('my_var', 'global', result)

    def test_ignoring_string_contents(self):
        code = "my_var = '('\nmy_"
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'global', result)

    def test_ignoring_comment_contents(self):
        code = "my_var = 10 #(\nmy_"
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'global', result)

    def test_ignoring_string_contents_backslash_plus_quotes(self):
        code = "my_var = '\\''\nmy_"
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'global', result)

    def test_ignoring_string_contents_backslash_plus_backslash(self):
        code = "my_var = '\\\\'\nmy_"
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'global', result)

    def test_not_proposing_later_defined_variables_in_current_block(self):
        code = "my_\nmy_var = 10\n"
        result = self._assist(code, 3, later_locals=False)
        self.assert_completion_not_in_result('my_var', 'global', result)

    def test_not_proposing_later_defined_variables_in_current_function(self):
        code = "def f():\n    my_\n    my_var = 10\n"
        result = self._assist(code, 16, later_locals=False)
        self.assert_completion_not_in_result('my_var', 'local', result)

    def test_ignoring_string_contents_with_triple_quotes(self):
        code = "my_var = '''(\n'('''\nmy_"
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'global', result)

    def test_ignoring_string_contents_with_triple_quotes_and_backslash(self):
        code = 'my_var = """\\"""("""\nmy_'
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'global', result)

    def test_ignor_str_contents_with_triple_quotes_and_double_backslash(self):
        code = 'my_var = """\\\\"""\nmy_'
        result = self._assist(code)
        self.assert_completion_in_result('my_var', 'global', result)

    def test_reporting_params_when_in_the_first_line_of_a_function(self):
        code = 'def f(param):\n    para'
        result = self._assist(code)
        self.assert_completion_in_result('param', 'local', result)

    def test_code_assist_when_having_a_two_line_function_header(self):
        code = 'def f(param1,\n      param2):\n    para'
        result = self._assist(code)
        self.assert_completion_in_result('param1', 'local', result)

    def test_code_assist_with_function_with_two_line_return(self):
        code = 'def f(param1, param2):\n    return(param1,\n           para'
        result = self._assist(code)
        self.assert_completion_in_result('param2', 'local', result)

    def test_get_definition_location(self):
        code = 'def a_func():\n    pass\na_func()'
        result = get_definition_location(self.project, code, len(code) - 3)
        self.assertEqual((None, 1), result)

    def test_get_definition_location_underlined_names(self):
        code = 'def a_sample_func():\n    pass\na_sample_func()'
        result = get_definition_location(self.project, code, len(code) - 11)
        self.assertEqual((None, 1), result)

    def test_get_definition_location_dotted_names_method(self):
        code = 'class AClass(object):\n' \
               '    @staticmethod\n' \
               '    def a_method():\n' \
               '        pass\n' \
               'AClass.a_method()'
        result = get_definition_location(self.project, code, len(code) - 3)
        self.assertEqual((None, 3), result)

    def test_get_definition_location_dotted_names_property(self):
        code = 'class AClass(object):\n' \
               '    @property\n' \
               '    @somedecorator\n' \
               '    def a_method():\n' \
               '        pass\n' \
               'AClass.a_method()'
        result = get_definition_location(self.project, code, len(code) - 3)
        self.assertEqual((None, 4), result)

    def test_get_definition_location_dotted_names_free_function(self):
        code = '@custom_decorator\n' \
               'def a_method():\n' \
               '    pass\n' \
               'a_method()'
        result = get_definition_location(self.project, code, len(code) - 3)
        self.assertEqual((None, 2), result)

    @testutils.only_for_versions_higher('3.5')
    def test_get_definition_location_dotted_names_async_def(self):
        code = 'class AClass(object):\n' \
               '    @property\n' \
               '    @decorator2\n' \
               '    async def a_method():\n' \
               '        pass\n' \
               'AClass.a_method()'
        result = get_definition_location(self.project, code, len(code) - 3)
        self.assertEqual((None, 4), result)

    def test_get_definition_location_dotted_names_class(self):
        code = '@custom_decorator\n' \
               'class AClass(object):\n' \
               '    def a_method():\n' \
               '        pass\n' \
               'AClass.a_method()'
        result = get_definition_location(self.project, code, len(code) - 12)
        self.assertEqual((None, 2), result)

    def test_get_definition_location_dotted_names_with_space(self):
        code = 'class AClass(object):\n' \
               '    @staticmethod\n' \
               '    def a_method():\n' \
               '        \n' \
               '        pass\n' \
               'AClass.a_method()'
        result = get_definition_location(self.project, code, len(code) - 3)
        self.assertEqual((None, 3), result)

    def test_get_definition_location_dotted_names_inline_body(self):
        code = 'class AClass(object):\n' \
               '    @staticmethod\n' \
               '    def a_method(): pass\n' \
               'AClass.a_method()'
        result = get_definition_location(self.project, code, len(code) - 3)
        self.assertEqual((None, 3), result)

    def test_get_definition_location_dotted_names_inline_body_split_arg(self):
        code = 'class AClass(object):\n' \
               '    @staticmethod\n' \
               '    def a_method(\n' \
               '        self,\n' \
               '        arg1\n' \
               '    ): pass\n' \
               'AClass.a_method()'
        result = get_definition_location(self.project, code, len(code) - 3)
        self.assertEqual((None, 3), result)

    def test_get_definition_location_dotted_module_names(self):
        module_resource = testutils.create_module(self.project, 'mod')
        module_resource.write('def a_func():\n    pass\n')
        code = 'import mod\nmod.a_func()'
        result = get_definition_location(self.project, code, len(code) - 3)
        self.assertEqual((module_resource, 1), result)

    def test_get_definition_location_for_nested_packages(self):
        mod1 = testutils.create_module(self.project, 'mod1')
        pkg1 = testutils.create_package(self.project, 'pkg1')
        pkg2 = testutils.create_package(self.project, 'pkg2', pkg1)
        mod1.write('import pkg1.pkg2.mod2')

        init_dot_py = pkg2.get_child('__init__.py')
        found_pyname = get_definition_location(self.project, mod1.read(),
                                               mod1.read().index('pkg2') + 1)
        self.assertEqual(init_dot_py, found_pyname[0])

    def test_get_definition_location_unknown(self):
        code = 'a_func()\n'
        result = get_definition_location(self.project, code, len(code) - 3)
        self.assertEqual((None, None), result)

    def test_get_definition_location_dot_spaces(self):
        code = 'class AClass(object):\n    ' \
               '@staticmethod\n    def a_method():\n' \
               '        pass\nAClass.\\\n     a_method()'
        result = get_definition_location(self.project, code, len(code) - 3)
        self.assertEqual((None, 3), result)

    def test_get_definition_location_dot_line_break_inside_parens(self):
        code = 'class A(object):\n    def a_method(self):\n        pass\n' + \
               '(A.\na_method)'
        result = get_definition_location(self.project, code,
                                         code.rindex('a_method') + 1)
        self.assertEqual((None, 2), result)

    def test_if_scopes_in_other_scopes_for_get_definition_location(self):
        code = 'def f(a_var):\n    pass\na_var = 10\n' + \
               'if True:\n' + \
               '    print(a_var)\n'
        result = get_definition_location(self.project, code, len(code) - 3)
        self.assertEqual((None, 3), result)

    def test_get_definition_location_false_triple_quoted_string(self):
        code = dedent('''\
            def foo():
                a = 0
                p = "foo"""

            def bar():
                a = 1
                a += 1
        ''')
        result = get_definition_location(self.project, code, code.index("a += 1"))
        self.assertEqual((None, 6), result)

    def test_code_assists_in_parens(self):
        code = 'def a_func(a_var):\n    pass\na_var = 10\na_func(a_'
        result = self._assist(code)
        self.assert_completion_in_result('a_var', 'global', result)

    def test_simple_type_inferencing(self):
        code = 'class Sample(object):\n' \
               '    def __init__(self, a_param):\n' \
               '        pass\n' \
               '    def a_method(self):\n' \
               '        pass\n' \
               'Sample("hey").a_'
        result = self._assist(code)
        self.assert_completion_in_result('a_method', 'attribute', result)

    def test_proposals_sorter(self):
        code = 'def my_sample_function(self):\n' + \
               '    my_sample_var = 20\n' + \
               '    my_sample_'
        proposals = sorted_proposals(self._assist(code))
        self.assertEqual('my_sample_var', proposals[0].name)
        self.assertEqual('my_sample_function', proposals[1].name)

    def test_proposals_sorter_for_methods_and_attributes(self):
        code = 'class A(object):\n' + \
               '    def __init__(self):\n' + \
               '        self.my_a_var = 10\n' + \
               '    def my_b_func(self):\n' + \
               '        pass\n' + \
               '    def my_c_func(self):\n' + \
               '        pass\n' + \
               'a_var = A()\n' + \
               'a_var.my_'
        proposals = sorted_proposals(self._assist(code))
        self.assertEqual('my_b_func', proposals[0].name)
        self.assertEqual('my_c_func', proposals[1].name)
        self.assertEqual('my_a_var', proposals[2].name)

    def test_proposals_sorter_for_global_methods_and_funcs(self):
        code = 'def my_b_func(self):\n' + \
               '    pass\n' + \
               'my_a_var = 10\n' + \
               'my_'
        proposals = sorted_proposals(self._assist(code))
        self.assertEqual('my_b_func', proposals[0].name)
        self.assertEqual('my_a_var', proposals[1].name)

    def test_proposals_sorter_underlined_methods(self):
        code = 'class A(object):\n' + \
               '    def _my_func(self):\n' + \
               '        self.my_a_var = 10\n' + \
               '    def my_func(self):\n' + \
               '        pass\n' + \
               'a_var = A()\n' + \
               'a_var.'
        proposals = sorted_proposals(self._assist(code))
        self.assertEqual('my_func', proposals[0].name)
        self.assertEqual('_my_func', proposals[1].name)

    def test_proposals_sorter_and_scope_prefs(self):
        code = 'my_global_var = 1\n' \
               'def func(self):\n' \
               '    my_local_var = 2\n' \
               '    my_'
        result = self._assist(code)
        proposals = sorted_proposals(result, scopepref=['global', 'local'])
        self.assertEqual('my_global_var', proposals[0].name)
        self.assertEqual('my_local_var', proposals[1].name)

    def test_proposals_sorter_and_type_prefs(self):
        code = 'my_global_var = 1\n' \
               'def my_global_func(self):\n' \
               '    pass\n' \
               'my_'
        result = self._assist(code)
        proposals = sorted_proposals(result, typepref=['instance', 'function'])
        self.assertEqual('my_global_var', proposals[0].name)
        self.assertEqual('my_global_func', proposals[1].name)

    def test_proposals_sorter_and_missing_type_in_typepref(self):
        code = 'my_global_var = 1\n' \
               'def my_global_func():\n' \
               '    pass\n' \
               'my_'
        result = self._assist(code)
        proposals = sorted_proposals(result, typepref=['function'])  # noqa

    def test_get_pydoc_unicode(self):
        src = u'# coding: utf-8\ndef foo():\n  u"юникод-объект"'
        doc = get_doc(self.project, src, src.index('foo') + 1)
        self.assertTrue(isinstance(doc, unicode))
        self.assertTrue(u'юникод-объект' in doc)

    def test_get_pydoc_utf8_bytestring(self):
        src = u'# coding: utf-8\ndef foo():\n  "байтстринг"'
        doc = get_doc(self.project, src, src.index('foo') + 1)
        self.assertTrue(isinstance(doc, unicode))
        self.assertTrue(u'байтстринг' in doc)

    def test_get_pydoc_for_functions(self):
        src = 'def a_func():\n' \
              '    """a function"""\n' \
              '    a_var = 10\n' \
              'a_func()'
        self.assertTrue(get_doc(self.project, src, len(src) - 4).
                        endswith('a function'))
        get_doc(self.project, src, len(src) - 4).index('a_func()')

    def test_get_pydoc_for_classes(self):
        src = 'class AClass(object):\n    pass\n'
        get_doc(self.project, src, src.index('AClass') + 1).index('AClass')

    def test_get_pydoc_for_classes_with_init(self):
        src = 'class AClass(object):\n    def __init__(self):\n        pass\n'
        get_doc(self.project, src, src.index('AClass') + 1).index('AClass')

    def test_get_pydoc_for_modules(self):
        mod = testutils.create_module(self.project, 'mod')
        mod.write('"""a module"""\n')
        src = 'import mod\nmod'
        self.assertEqual('a module', get_doc(self.project, src, len(src) - 1))

    def test_get_pydoc_for_builtins(self):
        src = 'print(object)\n'
        self.assertTrue(get_doc(self.project, src,
                                src.index('obj')) is not None)

    def test_get_pydoc_for_methods_should_include_class_name(self):
        src = 'class AClass(object):\n' \
              '    def a_method(self):\n'\
              '        """hey"""\n' \
              '        pass\n'
        doc = get_doc(self.project, src, src.index('a_method') + 1)
        doc.index('AClass.a_method')
        doc.index('hey')

    def test_get_pydoc_for_meths_should_inc_methods_from_super_classes(self):
        src = 'class A(object):\n' \
              '    def a_method(self):\n' \
              '        """hey1"""\n' \
              '        pass\n' \
              'class B(A):\n' \
              '    def a_method(self):\n' \
              '        """hey2"""\n' \
              '        pass\n'
        doc = get_doc(self.project, src, src.rindex('a_method') + 1)
        doc.index('A.a_method')
        doc.index('hey1')
        doc.index('B.a_method')
        doc.index('hey2')

    def test_get_pydoc_for_classes_should_name_super_classes(self):
        src = 'class A(object):\n    pass\n' \
              'class B(A):\n    pass\n'
        doc = get_doc(self.project, src, src.rindex('B') + 1)
        doc.index('B(A)')

    def test_get_pydoc_for_builtin_functions(self):
        src = 's = "hey"\ns.replace\n'
        doc = get_doc(self.project, src, src.rindex('replace') + 1)
        self.assertTrue(doc is not None)

    def test_commenting_errors_before_offset(self):
        src = 'lsjd lsjdf\ns = "hey"\ns.replace()\n'
        doc = get_doc(self.project, src, src.rindex('replace') + 1)  # noqa

    def test_proposing_variables_defined_till_the_end_of_scope(self):
        code = 'if True:\n    a_v\na_var = 10\n'
        result = self._assist(code, code.index('a_v') + 3)
        self.assert_completion_in_result('a_var', 'global', result)

    def test_completing_in_uncomplete_try_blocks(self):
        code = 'try:\n    a_var = 10\n    a_'
        result = self._assist(code)
        self.assert_completion_in_result('a_var', 'global', result)

    def test_completing_in_uncomplete_try_blocks_in_functions(self):
        code = 'def a_func():\n    try:\n        a_var = 10\n        a_'
        result = self._assist(code)
        self.assert_completion_in_result('a_var', 'local', result)

    def test_already_complete_try_blocks_with_finally(self):
        code = 'def a_func():\n    try:\n        a_var = 10\n        a_'
        result = self._assist(code)
        self.assert_completion_in_result('a_var', 'local', result)

    def test_already_complete_try_blocks_with_finally2(self):
        code = 'try:\n    a_var = 10\n    a_\nfinally:\n    pass\n'
        result = self._assist(code, code.rindex('a_') + 2)
        self.assert_completion_in_result('a_var', 'global', result)

    def test_already_complete_try_blocks_with_except(self):
        code = 'try:\n    a_var = 10\n    a_\nexcept Exception:\n    pass\n'
        result = self._assist(code, code.rindex('a_') + 2)
        self.assert_completion_in_result('a_var', 'global', result)

    def test_already_complete_try_blocks_with_except2(self):
        code = 'a_var = 10\ntry:\n    ' \
               'another_var = a_\n    another_var = 10\n' \
               'except Exception:\n    pass\n'
        result = self._assist(code, code.rindex('a_') + 2)
        self.assert_completion_in_result('a_var', 'global', result)

    def test_completing_ifs_in_uncomplete_try_blocks(self):
        code = 'try:\n    if True:\n        a_var = 10\n    a_'
        result = self._assist(code)
        self.assert_completion_in_result('a_var', 'global', result)

    def test_completing_ifs_in_uncomplete_try_blocks2(self):
        code = 'try:\n    if True:\n        a_var = 10\n        a_'
        result = self._assist(code)
        self.assert_completion_in_result('a_var', 'global', result)

    def test_completing_excepts_in_uncomplete_try_blocks(self):
        code = 'try:\n    pass\nexcept Exc'
        result = self._assist(code)
        self.assert_completion_in_result('Exception', 'builtin', result)

    def test_and_normal_complete_blocks_and_single_fixing(self):
        code = 'try:\n    range.\nexcept:\n    pass\n'
        result = self._assist(code, code.index('.'), maxfixes=1)  # noqa

    def test_nested_blocks(self):
        code = 'a_var = 10\ntry:\n    try:\n        a_v'
        result = self._assist(code)
        self.assert_completion_in_result('a_var', 'global', result)

    def test_proposing_function_keywords_when_calling(self):
        code = 'def f(p):\n    pass\nf(p'
        result = self._assist(code)
        self.assert_completion_in_result('p=', 'parameter_keyword', result)

    def test_proposing_function_keywords_when_calling_for_non_functions(self):
        code = 'f = 1\nf(p'
        result = self._assist(code)  # noqa

    def test_proposing_function_keywords_when_calling_extra_spaces(self):
        code = 'def f(p):\n    pass\nf( p'
        result = self._assist(code)
        self.assert_completion_in_result('p=', 'parameter_keyword', result)

    def test_proposing_function_keywords_when_calling_on_second_argument(self):
        code = 'def f(p1, p2):\n    pass\nf(1, p'
        result = self._assist(code)
        self.assert_completion_in_result('p2=', 'parameter_keyword', result)

    def test_proposing_function_keywords_when_calling_not_proposing_args(self):
        code = 'def f(p1, *args):\n    pass\nf(1, a'
        result = self._assist(code)
        self.assert_completion_not_in_result('args=', 'parameter_keyword',
                                             result)

    def test_propos_function_kwrds_when_call_with_no_noth_after_parens(self):
        code = 'def f(p):\n    pass\nf('
        result = self._assist(code)
        self.assert_completion_in_result('p=', 'parameter_keyword', result)

    def test_propos_function_kwrds_when_call_with_no_noth_after_parens2(self):
        code = 'def f(p):\n    pass\ndef g():\n    h = f\n    f('
        result = self._assist(code)
        self.assert_completion_in_result('p=', 'parameter_keyword', result)

    def test_codeassists_before_opening_of_parens(self):
        code = 'def f(p):\n    pass\na_var = 1\nf(1)\n'
        result = self._assist(code, code.rindex('f') + 1)
        self.assert_completion_not_in_result('a_var', 'global', result)

    def test_codeassist_before_single_line_indents(self):
        code = 'myvar = 1\nif True:\n    (myv\nif True:\n    pass\n'
        result = self._assist(code, code.rindex('myv') + 3)
        self.assert_completion_not_in_result('myvar', 'local', result)

    def test_codeassist_before_line_indents_in_a_blank_line(self):
        code = 'myvar = 1\nif True:\n    \nif True:\n    pass\n'
        result = self._assist(code, code.rindex('    ') + 4)
        self.assert_completion_not_in_result('myvar', 'local', result)

    def test_simple_get_calltips(self):
        src = 'def f():\n    pass\nvar = f()\n'
        doc = get_calltip(self.project, src, src.rindex('f'))
        self.assertEqual('f()', doc)

    def test_get_calltips_for_classes(self):
        src = 'class C(object):\n' \
              '    def __init__(self):\n        pass\nC('
        doc = get_calltip(self.project, src, len(src) - 1)
        self.assertEqual('C.__init__(self)', doc)

    def test_get_calltips_for_objects_with_call(self):
        src = 'class C(object):\n' \
              '    def __call__(self, p):\n        pass\n' \
              'c = C()\nc(1,'
        doc = get_calltip(self.project, src, src.rindex('c'))
        self.assertEqual('C.__call__(self, p)', doc)

    def test_get_calltips_and_including_module_name(self):
        src = 'class C(object):\n' \
              '    def __call__(self, p):\n        pass\n' \
              'c = C()\nc(1,'
        mod = testutils.create_module(self.project, 'mod')
        mod.write(src)
        doc = get_calltip(self.project, src, src.rindex('c'), mod)
        self.assertEqual('mod.C.__call__(self, p)', doc)

    def test_get_calltips_and_including_module_name_2(self):
        src = 'range()\n'
        doc = get_calltip(self.project, src, 1, ignore_unknown=True)
        self.assertTrue(doc is None)

    def test_removing_self_parameter(self):
        src = 'class C(object):\n' \
              '    def f(self):\n'\
              '        pass\n' \
              'C().f()'
        doc = get_calltip(self.project, src, src.rindex('f'), remove_self=True)
        self.assertEqual('C.f()', doc)

    def test_removing_self_parameter_and_more_than_one_parameter(self):
        src = 'class C(object):\n' \
              '    def f(self, p1):\n'\
              '        pass\n' \
              'C().f()'
        doc = get_calltip(self.project, src, src.rindex('f'), remove_self=True)
        self.assertEqual('C.f(p1)', doc)

    def test_lambda_calltip(self):
        src = 'foo = lambda x, y=1: None\n' \
              'foo()'
        doc = get_calltip(self.project, src, src.rindex('f'))
        self.assertEqual(doc, 'lambda(x, y)')

    def test_keyword_before_parens(self):
        code = 'if (1).:\n pass'
        result = self._assist(code, offset=len('if (1).'))
        self.assertTrue(result)

    # TESTING PROPOSAL'S KINDS AND TYPES.
    # SEE RELATION MATRIX IN `CompletionProposal`'s DOCSTRING

    def test_local_variable_completion_proposal(self):
        code = 'def foo():\n  xvar = 5\n  x'
        result = self._assist(code)
        self.assert_completion_in_result('xvar', 'local', result, 'instance')

    def test_global_variable_completion_proposal(self):
        code = 'yvar = 5\ny'
        result = self._assist(code)
        self.assert_completion_in_result('yvar', 'global', result, 'instance')

    def test_builtin_variable_completion_proposal(self):
        for varname in ('False', 'True'):
            result = self._assist(varname[0])
            self.assert_completion_in_result(varname, 'builtin', result,
                                             type='instance')

    def test_attribute_variable_completion_proposal(self):
        code = 'class AClass(object):\n  def foo(self):\n    ' \
               'self.bar = 1\n    self.b'
        result = self._assist(code)
        self.assert_completion_in_result('bar', 'attribute', result,
                                         type='instance')

    def test_local_class_completion_proposal(self):
        code = 'def foo():\n  class LocalClass(object): pass\n  Lo'
        result = self._assist(code)
        self.assert_completion_in_result('LocalClass', 'local', result,
                                         type='class')

    def test_global_class_completion_proposal(self):
        code = 'class GlobalClass(object): pass\nGl'
        result = self._assist(code)
        self.assert_completion_in_result('GlobalClass', 'global', result,
                                         type='class')

    def test_builtin_class_completion_proposal(self):
        for varname in ('object', 'dict', 'file'):
            result = self._assist(varname[0])
            self.assert_completion_in_result(varname, 'builtin', result,
                                             type='class')

    def test_attribute_class_completion_proposal(self):
        code = 'class Outer(object):\n  class Inner(object): pass\nOuter.'
        result = self._assist(code)
        self.assert_completion_in_result('Inner', 'attribute', result,
                                         type='class')

    def test_local_function_completion_proposal(self):
        code = 'def outer():\n  def inner(): pass\n  in'
        result = self._assist(code)
        self.assert_completion_in_result('inner', 'local', result,
                                         type='function')

    def test_global_function_completion_proposal(self):
        code = 'def foo(): pass\nf'
        result = self._assist(code)
        self.assert_completion_in_result('foo', 'global', result,
                                         type='function')

    def test_builtin_function_completion_proposal(self):
        code = 'a'
        result = self._assist(code)
        for expected in ('all', 'any', 'abs'):
            self.assert_completion_in_result(expected, 'builtin', result,
                                             type='function')

    def test_attribute_function_completion_proposal(self):
        code = 'class Some(object):\n  def method(self):\n    self.'
        result = self._assist(code)
        self.assert_completion_in_result('method', 'attribute', result,
                                         type='function')

    def test_local_module_completion_proposal(self):
        code = 'def foo():\n  import types\n  t'
        result = self._assist(code)
        self.assert_completion_in_result('types', 'imported', result,
                                         type='module')

    def test_global_module_completion_proposal(self):
        code = 'import operator\no'
        result = self._assist(code)
        self.assert_completion_in_result('operator', 'imported', result,
                                         type='module')

    def test_attribute_module_completion_proposal(self):
        code = 'class Some(object):\n  import os\nSome.o'
        result = self._assist(code)
        self.assert_completion_in_result('os', 'imported', result,
                                         type='module')

    def test_builtin_exception_completion_proposal(self):
        code = 'def blah():\n  Z'
        result = self._assist(code)
        self.assert_completion_in_result('ZeroDivisionError', 'builtin',
                                         result, type='class')

    def test_keyword_completion_proposal(self):
        code = 'f'
        result = self._assist(code)
        self.assert_completion_in_result('for', 'keyword', result, type=None)
        self.assert_completion_in_result('from', 'keyword', result, type=None)

    def test_parameter_keyword_completion_proposal(self):
        code = 'def func(abc, aloha, alpha, amigo): pass\nfunc(a'
        result = self._assist(code)
        for expected in ('abc=', 'aloha=', 'alpha=', 'amigo='):
            self.assert_completion_in_result(expected, 'parameter_keyword',
                                             result, type=None)

    def test_object_path_global(self):
        code = 'GLOBAL_VARIABLE = 42\n'
        resource = testutils.create_module(self.project, 'mod')
        resource.write(code)
        result = get_canonical_path(self.project, resource, 1)
        mod_path = os.path.join(self.project.address, 'mod.py')
        self.assertEqual(
            result, [(mod_path, 'MODULE'),
                     ('GLOBAL_VARIABLE', 'VARIABLE')])

    def test_object_path_attribute(self):
        code = 'class Foo(object):\n' \
               '    attr = 42\n'
        resource = testutils.create_module(self.project, 'mod')
        resource.write(code)
        result = get_canonical_path(self.project, resource, 24)
        mod_path = os.path.join(self.project.address, 'mod.py')
        self.assertEqual(
            result, [(mod_path, 'MODULE'), ('Foo', 'CLASS'),
                     ('attr', 'VARIABLE')])

    def test_object_path_subclass(self):
        code = 'class Foo(object):\n' \
               '    class Bar(object):\n' \
               '        pass\n'
        resource = testutils.create_module(self.project, 'mod')
        resource.write(code)
        result = get_canonical_path(self.project, resource, 30)
        mod_path = os.path.join(self.project.address, 'mod.py')
        self.assertEqual(
            result, [(mod_path, 'MODULE'), ('Foo', 'CLASS'),
                     ('Bar', 'CLASS')])

    def test_object_path_method_parameter(self):
        code = 'class Foo(object):\n' \
               '    def bar(self, a, b, c):\n' \
               '        pass\n'
        resource = testutils.create_module(self.project, 'mod')
        resource.write(code)
        result = get_canonical_path(self.project, resource, 41)
        mod_path = os.path.join(self.project.address, 'mod.py')
        self.assertEqual(
            result, [(mod_path, 'MODULE'), ('Foo', 'CLASS'),
                     ('bar', 'FUNCTION'), ('b', 'PARAMETER')])

    def test_object_path_variable(self):
        code = 'def bar(a):\n' \
               '    x = a + 42\n'
        resource = testutils.create_module(self.project, 'mod')
        resource.write(code)
        result = get_canonical_path(self.project, resource, 17)
        mod_path = os.path.join(self.project.address, 'mod.py')
        self.assertEqual(
            result, [(mod_path, 'MODULE'), ('bar', 'FUNCTION'),
                     ('x', 'VARIABLE')])


class CodeAssistInProjectsTest(unittest.TestCase):
    def setUp(self):
        super(CodeAssistInProjectsTest, self).setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore
        samplemod = testutils.create_module(self.project, 'samplemod')
        code = 'class SampleClass(object):\n' \
               '    def sample_method():\n        pass\n\n' \
               'def sample_func():\n    pass\n' \
               'sample_var = 10\n\n' \
               'def _underlined_func():\n    pass\n\n'
        samplemod.write(code)
        package = testutils.create_package(self.project, 'package')
        nestedmod = testutils.create_module(self.project,  # noqa
                                            'nestedmod', package)

    def tearDown(self):
        testutils.remove_project(self.project)
        super(self.__class__, self).tearDown()

    def _assist(self, code, resource=None, **kwds):
        return code_assist(self.project, code, len(code), resource, **kwds)

    def assert_completion_in_result(self, name, scope, result):
        for proposal in result:
            if proposal.name == name and proposal.scope == scope:
                return
        self.fail('completion <%s> not proposed' % name)

    def assert_completion_not_in_result(self, name, scope, result):
        for proposal in result:
            if proposal.name == name and proposal.scope == scope:
                self.fail('completion <%s> was proposed' % name)

    def test_simple_import(self):
        code = 'import samplemod\nsample'
        result = self._assist(code)
        self.assert_completion_in_result('samplemod', 'imported', result)

    def test_from_import_class(self):
        code = 'from samplemod import SampleClass\nSample'
        result = self._assist(code)
        self.assert_completion_in_result('SampleClass', 'imported', result)

    def test_from_import_function(self):
        code = 'from samplemod import sample_func\nsample'
        result = self._assist(code)
        self.assert_completion_in_result('sample_func', 'imported', result)

    def test_from_import_variable(self):
        code = 'from samplemod import sample_var\nsample'
        result = self._assist(code)
        self.assert_completion_in_result('sample_var', 'imported', result)

    def test_from_imports_inside_functions(self):
        code = 'def f():\n    from samplemod import SampleClass\n    Sample'
        result = self._assist(code)
        self.assert_completion_in_result('SampleClass', 'imported', result)

    def test_from_import_only_imports_imported(self):
        code = 'from samplemod import sample_func\nSample'
        result = self._assist(code)
        self.assert_completion_not_in_result('SampleClass', 'global', result)

    def test_from_import_star(self):
        code = 'from samplemod import *\nSample'
        result = self._assist(code)
        self.assert_completion_in_result('SampleClass', 'imported', result)

    def test_from_import_star2(self):
        code = 'from samplemod import *\nsample'
        result = self._assist(code)
        self.assert_completion_in_result('sample_func', 'imported', result)
        self.assert_completion_in_result('sample_var', 'imported', result)

    def test_from_import_star_not_imporing_underlined(self):
        code = 'from samplemod import *\n_under'
        result = self._assist(code)
        self.assert_completion_not_in_result('_underlined_func', 'global',
                                             result)

    def test_from_package_import_mod(self):
        code = 'from package import nestedmod\nnest'
        result = self._assist(code)
        self.assert_completion_in_result('nestedmod', 'imported', result)

    def test_completing_after_dot(self):
        code = 'class SampleClass(object):\n' \
               '    def sample_method(self):\n' \
               '        pass\n' \
               'SampleClass.sam'
        result = self._assist(code)
        self.assert_completion_in_result('sample_method', 'attribute', result)

    def test_completing_after_multiple_dots(self):
        code = 'class Class1(object):\n' \
               '    class Class2(object):\n' \
               '        def sample_method(self):\n' \
               '            pass\n' \
               'Class1.Class2.sam'
        result = self._assist(code)
        self.assert_completion_in_result('sample_method', 'attribute', result)

    def test_completing_after_self_dot(self):
        code = 'class Sample(object):\n' \
               '    def method1(self):\n' \
               '        pass\n' \
               '    def method2(self):\n' \
               '        self.m'
        result = self._assist(code)
        self.assert_completion_in_result('method1', 'attribute', result)

    def test_result_start_offset_for_dotted_completions(self):
        code = 'class Sample(object):\n' \
               '    def method1(self):\n' \
               '        pass\n' \
               'Sample.me'
        self.assertEqual(len(code) - 2, starting_offset(code, len(code)))

    def test_backslash_after_dots(self):
        code = 'class Sample(object):\n' \
               '    def a_method(self):\n' \
               '        pass\n' \
               'Sample.\\\n       a_m'
        result = self._assist(code)
        self.assert_completion_in_result('a_method', 'attribute', result)

    def test_not_proposing_global_names_after_dot(self):
        code = 'class Sample(object):\n' \
               '    def a_method(self):\n' \
               '        pass\n' \
               'Sample.'
        result = self._assist(code)
        self.assert_completion_not_in_result('Sample', 'global', result)

    def test_assist_on_relative_imports(self):
        pkg = testutils.create_package(self.project, 'pkg')
        mod1 = testutils.create_module(self.project, 'mod1', pkg)
        mod2 = testutils.create_module(self.project, 'mod2', pkg)
        mod1.write('def a_func():\n    pass\n')
        code = 'import mod1\nmod1.'
        result = self._assist(code, resource=mod2)
        self.assert_completion_in_result('a_func', 'imported', result)

    def test_get_location_on_relative_imports(self):
        pkg = testutils.create_package(self.project, 'pkg')
        mod1 = testutils.create_module(self.project, 'mod1', pkg)
        mod2 = testutils.create_module(self.project, 'mod2', pkg)
        mod1.write('def a_func():\n    pass\n')
        code = 'import mod1\nmod1.a_func\n'
        result = get_definition_location(self.project, code,
                                         len(code) - 2, mod2)
        self.assertEqual((mod1, 1), result)

    def test_get_definition_location_for_builtins(self):
        code = 'import sys\n'
        result = get_definition_location(self.project, code,
                                         len(code) - 2)
        self.assertEqual((None, None), result)

    def test_get_doc_on_relative_imports(self):
        pkg = testutils.create_package(self.project, 'pkg')
        mod1 = testutils.create_module(self.project, 'mod1', pkg)
        mod2 = testutils.create_module(self.project, 'mod2', pkg)
        mod1.write('def a_func():\n    """hey"""\n    pass\n')
        code = 'import mod1\nmod1.a_func\n'
        result = get_doc(self.project, code, len(code) - 2, mod2)
        self.assertTrue(result.endswith('hey'))

    def test_get_doc_on_from_import_module(self):
        mod1 = testutils.create_module(self.project, 'mod1')
        mod1.write('"""mod1 docs"""\nvar = 1\n')
        code = 'from mod1 import var\n'
        result = get_doc(self.project, code, code.index('mod1'))
        result.index('mod1 docs')

    def test_fixing_errors_with_maxfixes_in_resources(self):
        mod = testutils.create_module(self.project, 'mod')
        code = 'def f():\n    sldj sldj\ndef g():\n    ran'
        mod.write(code)
        result = self._assist(code, maxfixes=2, resource=mod)
        self.assertTrue(len(result) > 0)

    def test_completing_names_after_from_import(self):
        mod1 = testutils.create_module(self.project, 'mod1')
        mod2 = testutils.create_module(self.project, 'mod2')
        mod1.write('myvar = None\n')
        result = self._assist('from mod1 import myva', resource=mod2)
        self.assertTrue(len(result) > 0)
        self.assert_completion_in_result('myvar', 'global', result)

    def test_completing_names_after_from_import_and_sorted_proposals(self):
        mod1 = testutils.create_module(self.project, 'mod1')
        mod2 = testutils.create_module(self.project, 'mod2')
        mod1.write('myvar = None\n')
        result = self._assist('from mod1 import myva', resource=mod2)
        result = sorted_proposals(result)
        self.assertTrue(len(result) > 0)
        self.assert_completion_in_result('myvar', 'global', result)

    def test_completing_names_after_from_import2(self):
        mod1 = testutils.create_module(self.project, 'mod1')
        mod2 = testutils.create_module(self.project, 'mod2')
        mod1.write('myvar = None\n')
        result = self._assist('from mod1 import ', resource=mod2)
        self.assertTrue(len(result) > 0)
        self.assert_completion_in_result('myvar', 'global', result)

    def test_starting_expression(self):
        code = 'l = list()\nl.app'
        self.assertEqual('l.app', starting_expression(code, len(code)))


def suite():
    result = unittest.TestSuite()
    result.addTests(unittest.makeSuite(CodeAssistTest))
    result.addTests(unittest.makeSuite(CodeAssistInProjectsTest))
    return result

if __name__ == '__main__':
    unittest.main()



try:
    import unittest2 as unittest
except ImportError:
    import unittest

from rope.base import builtins, libutils, pyobjects
from ropetest import testutils
from rope.base.builtins import Dict

class BuiltinTypesTest(unittest.TestCase):

    def setUp(self):
        super(BuiltinTypesTest, self).setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore
        self.mod = testutils.create_module(self.project, 'mod')

    def tearDown(self):
        testutils.remove_project(self.project)
        super(BuiltinTypesTest, self).tearDown()

    def test_simple_case(self):
        self.mod.write('l = []\n')
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue('append' in pymod['l'].get_object())

    def test_holding_type_information(self):
        self.mod.write('class C(object):\n    pass\n'
                       'l = [C()]\na_var = l.pop()\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_get_items(self):
        self.mod.write('class C(object):'
                       '\n    def __getitem__(self, i):\n        return C()\n'
                       'c = C()\na_var = c[0]')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_get_items_for_lists(self):
        self.mod.write('class C(object):\n    pass\nl = [C()]\na_var = l[0]\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_get_items_from_slices(self):
        self.mod.write('class C(object):\n    pass'
                       '\nl = [C()]\na_var = l[:].pop()\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_simple_for_loops(self):
        self.mod.write('class C(object):\n    pass\nl = [C()]\n'
                       'for c in l:\n    a_var = c\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_definition_location_for_loop_variables(self):
        self.mod.write('class C(object):\n    pass\nl = [C()]\n'
                       'for c in l:\n    pass\n')
        pymod = self.project.get_pymodule(self.mod)
        c_var = pymod['c']
        self.assertEqual((pymod, 4), c_var.get_definition_location())

    def test_simple_case_for_dicts(self):
        self.mod.write('d = {}\n')
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue('get' in pymod['d'].get_object())

    def test_get_item_for_dicts(self):
        self.mod.write('class C(object):\n    pass\n'
                       'd = {1: C()}\na_var = d[1]\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_dict_function_parent(self):
        self.mod.write('d = {1: 2}\n'
                       'a_var = d.keys()')
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod['d'].get_object()['keys'].get_object()
        self.assertEqual(type(a_var.parent), Dict)

    def test_popping_dicts(self):
        self.mod.write('class C(object):\n    pass\n'
                       'd = {1: C()}\na_var = d.pop(1)\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_getting_keys_from_dicts(self):
        self.mod.write('class C1(object):\n    pass\n'
                       'class C2(object):\n    pass\n'
                       'd = {C1(): C2()}\nfor c in d.keys():\n    a_var = c\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C1'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_getting_values_from_dicts(self):
        self.mod.write('class C1(object):\n    pass\n'
                       'class C2(object):\n    pass\n'
                       'd = {C1(): C2()}\nfor c in d.values():'
                       '\n    a_var = c\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C2'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_getting_iterkeys_from_dicts(self):
        self.mod.write('class C1(object):\n    pass'
                       '\nclass C2(object):\n    pass\n'
                       'd = {C1(): C2()}\nfor c in d.keys():\n    a_var = c\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C1'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_getting_itervalues_from_dicts(self):
        self.mod.write('class C1(object):\n    pass'
                       '\nclass C2(object):\n    pass\n'
                       'd = {C1(): C2()}\nfor c in d.values():'
                       '\n    a_var = c\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C2'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_using_copy_for_dicts(self):
        self.mod.write('class C1(object):\n    pass'
                       '\nclass C2(object):\n    pass\n'
                       'd = {C1(): C2()}\nfor c in d.copy():\n    a_var = c\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C1'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_tuple_assignments_for_items(self):
        self.mod.write('class C1(object):\n    pass'
                       '\nclass C2(object):\n    pass\n'
                       'd = {C1(): C2()}\nkey, value = d.items()[0]\n')
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod['C1'].get_object()
        c2_class = pymod['C2'].get_object()
        key = pymod['key'].get_object()
        value = pymod['value'].get_object()
        self.assertEqual(c1_class, key.get_type())
        self.assertEqual(c2_class, value.get_type())

    def test_tuple_assignment_for_lists(self):
        self.mod.write('class C(object):\n    pass\n'
                       'l = [C(), C()]\na, b = l\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a'].get_object()
        b_var = pymod['b'].get_object()
        self.assertEqual(c_class, a_var.get_type())
        self.assertEqual(c_class, b_var.get_type())

    def test_tuple_assignments_for_iteritems_in_fors(self):
        self.mod.write('class C1(object):\n    pass\n'
                       'class C2(object):\n    pass\n'
                       'd = {C1(): C2()}\n'
                       'for x, y in d.items():\n    a = x;\n    b = y\n')
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod['C1'].get_object()
        c2_class = pymod['C2'].get_object()
        a_var = pymod['a'].get_object()
        b_var = pymod['b'].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_simple_tuple_assignments(self):
        self.mod.write('class C1(object):'
                       '\n    pass\nclass C2(object):\n    pass\n'
                       'a, b = C1(), C2()\n')
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod['C1'].get_object()
        c2_class = pymod['C2'].get_object()
        a_var = pymod['a'].get_object()
        b_var = pymod['b'].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_overriding_builtin_names(self):
        self.mod.write('class C(object):\n    pass\nlist = C\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        list_var = pymod['list'].get_object()
        self.assertEqual(c_class, list_var)

    def test_simple_builtin_scope_test(self):
        self.mod.write('l = list()\n')
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue('append' in pymod['l'].get_object())

    def test_simple_sets(self):
        self.mod.write('s = set()\n')
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue('add' in pymod['s'].get_object())

    def test_making_lists_using_the_passed_argument_to_init(self):
        self.mod.write('class C(object):\n    pass\nl1 = [C()]\n'
                       'l2 = list(l1)\na_var = l2.pop()')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_making_tuples_using_the_passed_argument_to_init(self):
        self.mod.write('class C(object):\n    pass\nl1 = [C()]\n'
                       'l2 = tuple(l1)\na_var = l2[0]')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_making_sets_using_the_passed_argument_to_init(self):
        self.mod.write('class C(object):\n    pass\nl1 = [C()]\n'
                       'l2 = set(l1)\na_var = l2.pop()')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_making_dicts_using_the_passed_argument_to_init(self):
        self.mod.write('class C1(object):\n    pass\n'
                       'class C2(object):\n    pass\n'
                       'l1 = [(C1(), C2())]\n'
                       'l2 = dict(l1)\na, b = l2.items()[0]')
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod['C1'].get_object()
        c2_class = pymod['C2'].get_object()
        a_var = pymod['a'].get_object()
        b_var = pymod['b'].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_range_builtin_function(self):
        self.mod.write('l = range(1)\n')
        pymod = self.project.get_pymodule(self.mod)
        l = pymod['l'].get_object()
        self.assertTrue('append' in l)

    def test_reversed_builtin_function(self):
        self.mod.write('class C(object):\n    pass\nl = [C()]\n'
                       'for x in reversed(l):\n    a_var = x\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_sorted_builtin_function(self):
        self.mod.write('class C(object):\n    pass\nl = [C()]\n'
                       'a_var = sorted(l).pop()\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_super_builtin_function(self):
        self.mod.write(
            'class C(object):\n    pass\n'
            'class A(object):\n    def a_f(self):\n        return C()\n'
            'class B(A):\n    def b_f(self):\n        '
            'return super(B, self).a_f()\n'
            'a_var = B.b_f()\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_file_builtin_type(self):
        self.mod.write('for line in open("file.txt"):\n    a_var = line\n')
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod['a_var'].get_object()
        self.assertTrue(isinstance(a_var.get_type(), builtins.Str))

    def test_property_builtin_type(self):
        self.mod.write('p = property()\n')
        pymod = self.project.get_pymodule(self.mod)
        p_var = pymod['p'].get_object()
        self.assertTrue('fget' in p_var)

    def test_lambda_functions(self):
        self.mod.write('l = lambda: 1\n')
        pymod = self.project.get_pymodule(self.mod)
        l_var = pymod['l'].get_object()
        self.assertEqual(pyobjects.get_base_type('Function'),
                          l_var.get_type())

    def test_lambda_function_definition(self):
        self.mod.write('l = lambda x, y = 2, *a, **b: x + y\n')
        pymod = self.project.get_pymodule(self.mod)
        l_var = pymod['l'].get_object()
        self.assertTrue(l_var.get_name() is not None)
        self.assertEqual(len(l_var.get_param_names()), 4)
        self.assertEqual((pymod, 1),
                          pymod['l'].get_definition_location())

    def test_lambdas_that_return_unknown(self):
        self.mod.write('a_var = (lambda: None)()\n')
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod['a_var'].get_object()
        self.assertTrue(a_var is not None)

    def test_builtin_zip_function(self):
        self.mod.write(
            'class C1(object):\n    pass\nclass C2(object):\n    pass\n'
            'c1_list = [C1()]\nc2_list = [C2()]\n'
            'a, b = zip(c1_list, c2_list)[0]')
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod['C1'].get_object()
        c2_class = pymod['C2'].get_object()
        a_var = pymod['a'].get_object()
        b_var = pymod['b'].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_builtin_zip_function_with_more_than_two_args(self):
        self.mod.write(
            'class C1(object):\n    pass\nclass C2(object):\n    pass\n'
            'c1_list = [C1()]\nc2_list = [C2()]\n'
            'a, b, c = zip(c1_list, c2_list, c1_list)[0]')
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod['C1'].get_object()
        c2_class = pymod['C2'].get_object()
        a_var = pymod['a'].get_object()
        b_var = pymod['b'].get_object()
        c_var = pymod['c'].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())
        self.assertEqual(c1_class, c_var.get_type())

    def test_wrong_arguments_to_zip_function(self):
        self.mod.write(
            'class C1(object):\n    pass\nc1_list = [C1()]\n'
            'a, b = zip(c1_list, 1)[0]')
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod['C1'].get_object()
        a_var = pymod['a'].get_object()
        b_var = pymod['b'].get_object()  # noqa
        self.assertEqual(c1_class, a_var.get_type())

    def test_enumerate_builtin_function(self):
        self.mod.write('class C(object):\n    pass\nl = [C()]\n'
                       'for i, x in enumerate(l):\n    a_var = x\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_builtin_class_get_name(self):
        self.assertEqual('object',
                          builtins.builtins['object'].get_object().get_name())
        self.assertEqual(
            'property', builtins.builtins['property'].get_object().get_name())

    def test_star_args_and_double_star_args(self):
        self.mod.write('def func(p, *args, **kwds):\n    pass\n')
        pymod = self.project.get_pymodule(self.mod)
        func_scope = pymod['func'].get_object().get_scope()
        args = func_scope['args'].get_object()
        kwds = func_scope['kwds'].get_object()
        self.assertTrue(isinstance(args.get_type(), builtins.List))
        self.assertTrue(isinstance(kwds.get_type(), builtins.Dict))

    def test_simple_list_comprehension_test(self):
        self.mod.write('a_var = [i for i in range(10)]\n')
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod['a_var'].get_object()
        self.assertTrue(isinstance(a_var.get_type(), builtins.List))

    def test_simple_list_generator_expression(self):
        self.mod.write('a_var = (i for i in range(10))\n')
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod['a_var'].get_object()
        self.assertTrue(isinstance(a_var.get_type(), builtins.Iterator))

    def test_iter_builtin_function(self):
        self.mod.write('class C(object):\n    pass\nl = [C()]\n'
                       'for c in iter(l):\n    a_var = c\n')
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod['C'].get_object()
        a_var = pymod['a_var'].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_simple_int_type(self):
        self.mod.write('l = 1\n')
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(builtins.builtins['int'].get_object(),
                          pymod['l'].get_object().get_type())

    def test_simple_float_type(self):
        self.mod.write('l = 1.0\n')
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(builtins.builtins['float'].get_object(),
                          pymod['l'].get_object().get_type())

    def test_simple_float_type2(self):
        self.mod.write('l = 1e1\n')
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(builtins.builtins['float'].get_object(),
                          pymod['l'].get_object().get_type())

    def test_simple_complex_type(self):
        self.mod.write('l = 1.0j\n')
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(builtins.builtins['complex'].get_object(),
                          pymod['l'].get_object().get_type())

    def test_handling_unaryop_on_ints(self):
        self.mod.write('l = -(1)\n')
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(builtins.builtins['int'].get_object(),
                          pymod['l'].get_object().get_type())

    def test_handling_binop_on_ints(self):
        self.mod.write('l = 1 + 1\n')
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(builtins.builtins['int'].get_object(),
                          pymod['l'].get_object().get_type())

    def test_handling_compares(self):
        self.mod.write('l = 1 == 1\n')
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(builtins.builtins['bool'].get_object(),
                          pymod['l'].get_object().get_type())

    def test_handling_boolops(self):
        self.mod.write('l = 1 and 2\n')
        pymod = self.project.get_pymodule(self.mod)
        self.assertEqual(builtins.builtins['int'].get_object(),
                          pymod['l'].get_object().get_type())

    def test_binary_or_left_value_unknown(self):
        code = 'var = (asdsd or 3)\n'
        pymod = libutils.get_string_module(self.project, code)
        self.assertEqual(builtins.builtins['int'].get_object(),
                          pymod['var'].get_object().get_type())

    def test_unknown_return_object(self):
        src = 'import sys\n' \
              'def foo():\n' \
              '  res = set(sys.builtin_module_names)\n' \
              '  if foo: res.add(bar)\n'
        self.project.prefs['import_dynload_stdmods'] = True
        self.mod.write(src)
        self.project.pycore.analyze_module(self.mod)

    def test_abstractmethods_attribute(self):
        # see http://bugs.python.org/issue10006 for details
        src = 'class SubType(type): pass\nsubtype = SubType()\n'
        self.mod.write(src)
        self.project.pycore.analyze_module(self.mod)


class BuiltinModulesTest(unittest.TestCase):

    def setUp(self):
        super(BuiltinModulesTest, self).setUp()
        self.project = testutils.sample_project(
            extension_modules=['time', 'invalid', 'invalid.sub'])
        self.mod = testutils.create_module(self.project, 'mod')

    def tearDown(self):
        testutils.remove_project(self.project)
        super(BuiltinModulesTest, self).tearDown()

    def test_simple_case(self):
        self.mod.write('import time')
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue('time' in pymod['time'].get_object())

    def test_ignored_extensions(self):
        self.mod.write('import os')
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue('rename' not in pymod['os'].get_object())

    def test_ignored_extensions_2(self):
        self.mod.write('import os')
        pymod = self.project.get_pymodule(self.mod)
        self.assertTrue('rename' not in pymod['os'].get_object())

    def test_nonexistent_modules(self):
        self.mod.write('import invalid')
        pymod = self.project.get_pymodule(self.mod)
        pymod['invalid'].get_object()

    def test_nonexistent_modules_2(self):
        self.mod.write('import invalid\nimport invalid.sub')
        pymod = self.project.get_pymodule(self.mod)
        invalid = pymod['invalid'].get_object()
        self.assertTrue('sub' in invalid)

    def test_time_in_std_mods(self):
        import rope.base.stdmods
        self.assertTrue('time' in rope.base.stdmods.standard_modules())

    def test_timemodule_normalizes_to_time(self):
        import rope.base.stdmods
        self.assertEqual(
            rope.base.stdmods.normalize_so_name('timemodule.so'), 'time')


def suite():
    result = unittest.TestSuite()
    result.addTests(unittest.makeSuite(BuiltinTypesTest))
    result.addTests(unittest.makeSuite(BuiltinModulesTest))
    return result


if __name__ == '__main__':
    unittest.main()



try:
    import unittest2 as unittest
except ImportError:
    import unittest

from rope.refactor import multiproject, rename, move
from ropetest import testutils


class MultiProjectRefactoringTest(unittest.TestCase):

    def setUp(self):
        super(MultiProjectRefactoringTest, self).setUp()
        self.project1 = testutils.sample_project(foldername='testproject1')
        self.project2 = testutils.sample_project(foldername='testproject2')
        self.mod1 = self.project1.root.create_file('mod1.py')
        self.other = self.project1.root.create_file('other.py')
        self.mod2 = self.project2.root.create_file('mod2.py')

    def tearDown(self):
        testutils.remove_project(self.project1)
        testutils.remove_project(self.project2)
        super(MultiProjectRefactoringTest, self).tearDown()

    def test_trivial_rename(self):
        self.mod1.write('var = 1\n')
        refactoring = multiproject.MultiProjectRefactoring(
            rename.Rename, [])
        renamer = refactoring(self.project1, self.mod1, 1)
        multiproject.perform(renamer.get_all_changes('newvar'))
        self.assertEqual('newvar = 1\n', self.mod1.read())

    def test_rename(self):
        self.mod1.write('var = 1\n')
        self.mod2.write('import mod1\nmyvar = mod1.var\n')
        refactoring = multiproject.MultiProjectRefactoring(
            rename.Rename, [self.project2])
        renamer = refactoring(self.project1, self.mod1, 1)
        multiproject.perform(renamer.get_all_changes('newvar'))
        self.assertEqual('newvar = 1\n', self.mod1.read())
        self.assertEqual('import mod1\nmyvar = mod1.newvar\n',
                          self.mod2.read())

    def test_move(self):
        self.mod1.write('def a_func():\n    pass\n')
        self.mod2.write('import mod1\nmyvar = mod1.a_func()\n')
        refactoring = multiproject.MultiProjectRefactoring(
            move.create_move, [self.project2])
        renamer = refactoring(self.project1, self.mod1,
                              self.mod1.read().index('_func'))
        multiproject.perform(renamer.get_all_changes(self.other))
        self.assertEqual('', self.mod1.read())
        self.assertEqual('def a_func():\n    pass\n', self.other.read())
        self.assertEqual('import other\nmyvar = other.a_func()\n',
                          self.mod2.read())

    def test_rename_from_the_project_not_containing_the_change(self):
        self.project2.get_prefs().add('python_path', self.project1.address)
        self.mod1.write('var = 1\n')
        self.mod2.write('import mod1\nmyvar = mod1.var\n')
        refactoring = multiproject.MultiProjectRefactoring(
            rename.Rename, [self.project1])
        renamer = refactoring(self.project2, self.mod2,
                              self.mod2.read().rindex('var'))
        multiproject.perform(renamer.get_all_changes('newvar'))
        self.assertEqual('newvar = 1\n', self.mod1.read())
        self.assertEqual('import mod1\nmyvar = mod1.newvar\n',
                          self.mod2.read())


if __name__ == '__main__':
    unittest.main()



from textwrap import dedent
try:
    import unittest2 as unittest
except ImportError:
    import unittest

import rope.base.evaluate
from rope.base import libutils
from rope.base import exceptions, worder, codeanalyze
from rope.base.codeanalyze import (SourceLinesAdapter,
                                   LogicalLineFinder, get_block_start)
from ropetest import testutils


class SourceLinesAdapterTest(unittest.TestCase):

    def setUp(self):
        super(SourceLinesAdapterTest, self).setUp()

    def tearDown(self):
        super(SourceLinesAdapterTest, self).tearDown()

    def test_source_lines_simple(self):
        to_lines = SourceLinesAdapter('line1\nline2\n')
        self.assertEqual('line1', to_lines.get_line(1))
        self.assertEqual('line2', to_lines.get_line(2))
        self.assertEqual('', to_lines.get_line(3))
        self.assertEqual(3, to_lines.length())

    def test_source_lines_get_line_number(self):
        to_lines = SourceLinesAdapter('line1\nline2\n')
        self.assertEqual(1, to_lines.get_line_number(0))
        self.assertEqual(1, to_lines.get_line_number(5))
        self.assertEqual(2, to_lines.get_line_number(7))
        self.assertEqual(3, to_lines.get_line_number(12))

    def test_source_lines_get_line_start(self):
        to_lines = SourceLinesAdapter('line1\nline2\n')
        self.assertEqual(0, to_lines.get_line_start(1))
        self.assertEqual(6, to_lines.get_line_start(2))
        self.assertEqual(12, to_lines.get_line_start(3))

    def test_source_lines_get_line_end(self):
        to_lines = SourceLinesAdapter('line1\nline2\n')
        self.assertEqual(5, to_lines.get_line_end(1))
        self.assertEqual(11, to_lines.get_line_end(2))
        self.assertEqual(12, to_lines.get_line_end(3))

    def test_source_lines_last_line_with_no_new_line(self):
        to_lines = SourceLinesAdapter('line1')
        self.assertEqual(1, to_lines.get_line_number(5))


class WordRangeFinderTest(unittest.TestCase):

    def setUp(self):
        super(WordRangeFinderTest, self).setUp()

    def tearDown(self):
        super(WordRangeFinderTest, self).tearDown()

    def _find_primary(self, code, offset):
        word_finder = worder.Worder(code)
        result = word_finder.get_primary_at(offset)
        return result

    def test_keyword_before_parens(self):
        code = 'if (a_var).an_attr:\n    pass\n'
        self.assertEqual('(a_var).an_attr',
                          self._find_primary(code, code.index(':')))

    def test_inside_parans(self):
        code = 'a_func(a_var)'
        self.assertEqual('a_var', self._find_primary(code, 10))

    def test_simple_names(self):
        code = 'a_var = 10'
        self.assertEqual('a_var', self._find_primary(code, 3))

    def test_function_calls(self):
        code = 'sample_function()'
        self.assertEqual('sample_function', self._find_primary(code, 10))

    def test_attribute_accesses(self):
        code = 'a_var.an_attr'
        self.assertEqual('a_var.an_attr', self._find_primary(code, 10))

    def test_word_finder_on_word_beginning(self):
        code = 'print(a_var)\n'
        word_finder = worder.Worder(code)
        result = word_finder.get_word_at(code.index('a_var'))
        self.assertEqual('a_var', result)

    def test_word_finder_on_primary_beginning(self):
        code = 'print(a_var)\n'
        result = self._find_primary(code, code.index('a_var'))
        self.assertEqual('a_var', result)

    def test_word_finder_on_word_ending(self):
        code = 'print(a_var)\n'
        word_finder = worder.Worder(code)
        result = word_finder.get_word_at(code.index('a_var') + 5)
        self.assertEqual('a_var', result)

    def test_word_finder_on_primary_ending(self):
        code = 'print(a_var)\n'
        result = self._find_primary(code, code.index('a_var') + 5)
        self.assertEqual('a_var', result)

    def test_word_finder_on_primaries_with_dots_inside_parens(self):
        code = '(a_var.\nattr)'
        result = self._find_primary(code, code.index('attr') + 1)
        self.assertEqual('a_var.\nattr', result)

    def test_word_finder_on_primary_like_keyword(self):
        code = 'is_keyword = False\n'
        result = self._find_primary(code, 1)
        self.assertEqual('is_keyword', result)

    def test_keyword_before_parens_no_space(self):
        code = 'if(a_var).an_attr:\n    pass\n'
        self.assertEqual('(a_var).an_attr',
                         self._find_primary(code, code.index(':')))

    def test_strings(self):
        code = '"a string".split()'
        self.assertEqual('"a string".split', self._find_primary(code, 14))

    def test_function_calls2(self):
        code = 'file("afile.txt").read()'
        self.assertEqual('file("afile.txt").read',
                         self._find_primary(code, 18))

    def test_parens(self):
        code = '("afile.txt").split()'
        self.assertEqual('("afile.txt").split', self._find_primary(code, 18))

    def test_function_with_no_param(self):
        code = 'AClass().a_func()'
        self.assertEqual('AClass().a_func', self._find_primary(code, 12))

    def test_function_with_multiple_param(self):
        code = 'AClass(a_param, another_param, "a string").a_func()'
        self.assertEqual('AClass(a_param, another_param, "a string").a_func',
                         self._find_primary(code, 44))

    def test_param_expressions(self):
        code = 'AClass(an_object.an_attr).a_func()'
        self.assertEqual('an_object.an_attr', self._find_primary(code, 20))

    def test_string_parens(self):
        code = 'a_func("(").an_attr'
        self.assertEqual('a_func("(").an_attr', self._find_primary(code, 16))

    def test_extra_spaces(self):
        code = 'a_func  (  "(" ) .   an_attr'
        self.assertEqual('a_func  (  "(" ) .   an_attr',
                         self._find_primary(code, 26))

    def test_relative_import(self):
        code = "from .module import smt"
        self.assertEqual('.module',
                         self._find_primary(code, 5))

    def test_functions_on_ending_parens(self):
        code = 'A()'
        self.assertEqual('A()', self._find_primary(code, 2))

    def test_splitted_statement(self):
        word_finder = worder.Worder('an_object.an_attr')
        self.assertEqual(('an_object', 'an_at', 10),
                         word_finder.get_splitted_primary_before(15))

    def test_empty_splitted_statement(self):
        word_finder = worder.Worder('an_attr')
        self.assertEqual(('', 'an_at', 0),
                          word_finder.get_splitted_primary_before(5))

    def test_empty_splitted_statement2(self):
        word_finder = worder.Worder('an_object.')
        self.assertEqual(('an_object', '', 10),
                          word_finder.get_splitted_primary_before(10))

    def test_empty_splitted_statement3(self):
        word_finder = worder.Worder('')
        self.assertEqual(('', '', 0),
                          word_finder.get_splitted_primary_before(0))

    def test_empty_splitted_statement4(self):
        word_finder = worder.Worder('a_var = ')
        self.assertEqual(('', '', 8),
                          word_finder.get_splitted_primary_before(8))

    def test_empty_splitted_statement5(self):
        word_finder = worder.Worder('a.')
        self.assertEqual(('a', '', 2),
                          word_finder.get_splitted_primary_before(2))

    def test_operators_inside_parens(self):
        code = '(a_var + another_var).reverse()'
        self.assertEqual('(a_var + another_var).reverse',
                          self._find_primary(code, 25))

    def test_dictionaries(self):
        code = 'print({1: "one", 2: "two"}.keys())'
        self.assertEqual('{1: "one", 2: "two"}.keys',
                          self._find_primary(code, 29))

    def test_following_parens(self):
        code = 'a_var = a_func()()'
        result = self._find_primary(code, code.index(')(') + 3)
        self.assertEqual('a_func()()', result)

    def test_comments_for_finding_statements(self):
        code = '# var2 . \n  var3'
        self.assertEqual('var3', self._find_primary(code, code.index('3')))

    def test_str_in_comments_for_finding_statements(self):
        code = '# "var2" . \n  var3'
        self.assertEqual('var3', self._find_primary(code, code.index('3')))

    def test_comments_for_finding_statements2(self):
        code = 'var1 + "# var2".\n  var3'
        self.assertEqual('var3', self._find_primary(code, 21))

    def test_comments_for_finding_statements3(self):
        code = '"" + # var2.\n  var3'
        self.assertEqual('var3', self._find_primary(code, 21))

    def test_import_statement_finding(self):
        code = 'import mod\na_var = 10\n'
        word_finder = worder.Worder(code)
        self.assertTrue(word_finder.is_import_statement(code.index('mod') + 1))
        self.assertFalse(word_finder.is_import_statement(
            code.index('a_var') + 1))

    def test_import_statement_finding2(self):
        code = 'import a.b.c.d\nresult = a.b.c.d.f()\n'
        word_finder = worder.Worder(code)
        self.assertFalse(word_finder.is_import_statement(code.rindex('d') + 1))

    def test_word_parens_range(self):
        code = 's = str()\ns.title()\n'
        word_finder = worder.Worder(code)
        result = word_finder.get_word_parens_range(code.rindex('()') - 1)
        self.assertEqual((len(code) - 3, len(code) - 1), result)

    def test_getting_primary_before_get_index(self):
        code = '\na = (b + c).d[0]()\n'
        result = self._find_primary(code, len(code) - 2)
        self.assertEqual('(b + c).d[0]()', result)

    def test_getting_primary_and_strings_at_the_end_of_line(self):
        code = 'f(\'\\\'\')\n'
        result = self._find_primary(code, len(code) - 1)  # noqa

    def test_getting_primary_and_not_crossing_newlines(self):
        code = '\na = (b + c)\n(4 + 1).x\n'
        result = self._find_primary(code, len(code) - 1)
        self.assertEqual('(4 + 1).x', result)

    # XXX: cancatenated string literals
    def xxx_test_getting_primary_cancatenating_strs(self):
        code = 's = "a"\n"b" "c"\n'
        result = self._find_primary(code, len(code) - 2)
        self.assertEqual('"b" "c"', result)

    def test_is_a_function_being_called_with_parens_on_next_line(self):
        code = 'func\n(1, 2)\n'
        word_finder = worder.Worder(code)
        self.assertFalse(word_finder.is_a_function_being_called(1))

    # XXX: handling triple quotes
    def xxx_test_triple_quotes(self):
        code = 's = """string"""\n'
        result = self._find_primary(code, len(code) - 1)
        self.assertEqual('"""string"""', result)

    def test_triple_quotes_spanning_multiple_lines(self):
        code = 's = """\\\nl1\nl2\n """\n'
        result = self._find_primary(code, len(code) - 2)
        self.assertEqual('"""\\\nl1\nl2\n """', result)

    def test_get_word_parens_range_and_string_literals(self):
        code = 'f(1, ")", 2)\n'
        word_finder = worder.Worder(code)
        result = word_finder.get_word_parens_range(0)
        self.assertEqual((1, len(code) - 1), result)

    def test_is_assigned_here_for_equality_test(self):
        code = 'a == 1\n'
        word_finder = worder.Worder(code)
        self.assertFalse(word_finder.is_assigned_here(0))

    def test_is_assigned_here_for_not_equal_test(self):
        code = 'a != 1\n'
        word_finder = worder.Worder(code)
        self.assertFalse(word_finder.is_assigned_here(0))

    # XXX: is_assigned_here should work for tuple assignments
    def xxx_test_is_assigned_here_for_tuple_assignment(self):
        code = 'a, b = (1, 2)\n'
        word_finder = worder.Worder(code)
        self.assertTrue(word_finder.is_assigned_here(0))

    def test_is_from_with_from_import_and_multiline_parens(self):
        code = 'from mod import \\\n  (f,\n  g, h)\n'
        word_finder = worder.Worder(code)
        self.assertTrue(word_finder.is_from_statement(code.rindex('g')))

    def test_is_from_with_from_import_and_line_breaks_in_the_middle(self):
        code = 'from mod import f,\\\n g\n'
        word_finder = worder.Worder(code)
        self.assertTrue(word_finder.is_from_statement(code.rindex('g')))

    def test_one_letter_function_keyword_arguments(self):
        code = 'f(p=1)\n'
        word_finder = worder.Worder(code)
        index = code.rindex('p')
        self.assertTrue(word_finder.is_function_keyword_parameter(index))

    def test_find_parens_start(self):
        code = 'f(p)\n'
        finder = worder.Worder(code)
        self.assertEqual(1, finder.find_parens_start_from_inside(2))

    def test_underlined_find_parens_start(self):
        code = 'f(p="")\n'
        finder = worder.Worder(code)
        self.assertEqual(1, finder._find_parens_start(len(code) - 2))

    def test_find_parens_start_with_multiple_entries(self):
        code = 'myfunc(p1, p2, p3\n'
        finder = worder.Worder(code)
        self.assertEqual(code.index('('),
                          finder.find_parens_start_from_inside(len(code) - 1))

    def test_find_parens_start_with_nested_parens(self):
        code = 'myfunc(p1, (p2, p3), p4\n'
        finder = worder.Worder(code)
        self.assertEqual(code.index('('),
                          finder.find_parens_start_from_inside(len(code) - 1))

    def test_find_parens_start_with_parens_in_strs(self):
        code = 'myfunc(p1, "(", p4\n'
        finder = worder.Worder(code)
        self.assertEqual(code.index('('),
                          finder.find_parens_start_from_inside(len(code) - 1))

    def test_find_parens_start_with_parens_in_strs_in_multiple_lines(self):
        code = 'myfunc  (\np1\n , \n "(" \n, \np4\n'
        finder = worder.Worder(code)
        self.assertEqual(code.index('('),
                          finder.find_parens_start_from_inside(len(code) - 1))

    def test_is_on_function_keyword(self):
        code = 'myfunc(va'
        finder = worder.Worder(code)
        self.assertTrue(finder.is_on_function_call_keyword(len(code) - 1))

    def test_get_word_range_with_fstring(self):
        code = 'auth = 8\nmy_var = f"some value {auth}"\nprint(auth)\nother_val = "some other"'
        finder = worder.Worder(code)
        self.assertEqual(finder.get_word_range(45), (45, 49))


class ScopeNameFinderTest(unittest.TestCase):

    def setUp(self):
        super(ScopeNameFinderTest, self).setUp()
        self.project = testutils.sample_project()

    def tearDown(self):
        testutils.remove_project(self.project)
        super(ScopeNameFinderTest, self).tearDown()

    # FIXME: in normal scopes the interpreter raises `UnboundLocalName`
    # exception, but not in class bodies
    def xxx_test_global_name_in_class_body(self):
        code = 'a_var = 10\nclass C(object):\n    a_var = a_var\n'
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        result = name_finder.get_pyname_at(len(code) - 3)
        self.assertEqual(scope['a_var'], result)

    def test_class_variable_attribute_in_class_body(self):
        code = 'a_var = 10\nclass C(object):\n    a_var = a_var\n'
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        a_var_pyname = scope['C'].get_object()['a_var']
        result = name_finder.get_pyname_at(len(code) - 12)
        self.assertEqual(a_var_pyname, result)

    def test_class_variable_attribute_in_class_body2(self):
        code = 'a_var = 10\nclass C(object):\n    a_var \\\n= a_var\n'
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        a_var_pyname = scope['C'].get_object()['a_var']
        result = name_finder.get_pyname_at(len(code) - 12)
        self.assertEqual(a_var_pyname, result)

    def test_class_method_attribute_in_class_body(self):
        code = 'class C(object):\n    def a_method(self):\n        pass\n'
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        a_method_pyname = scope['C'].get_object()['a_method']
        result = name_finder.get_pyname_at(code.index('a_method') + 2)
        self.assertEqual(a_method_pyname, result)

    def test_inner_class_attribute_in_class_body(self):
        code = 'class C(object):\n    class CC(object):\n        pass\n'
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        a_class_pyname = scope['C'].get_object()['CC']
        result = name_finder.get_pyname_at(code.index('CC') + 2)
        self.assertEqual(a_class_pyname, result)

    def test_class_method_in_class_body_but_not_indexed(self):
        code = 'class C(object):\n    def func(self, func):\n        pass\n'
        scope = libutils.get_string_scope(self.project, code)
        a_func_pyname = scope.get_scopes()[0].get_scopes()[0]['func']
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        result = name_finder.get_pyname_at(code.index(', func') + 3)
        self.assertEqual(a_func_pyname, result)

    def test_function_but_not_indexed(self):
        code = 'def a_func(a_func):\n    pass\n'
        scope = libutils.get_string_scope(self.project, code)
        a_func_pyname = scope['a_func']
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        result = name_finder.get_pyname_at(code.index('a_func') + 3)
        self.assertEqual(a_func_pyname, result)

    def test_modules_after_from_statements(self):
        root_folder = self.project.root
        mod = testutils.create_module(self.project, 'mod', root_folder)
        mod.write('def a_func():\n    pass\n')
        code = 'from mod import a_func\n'
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        mod_pyobject = self.project.get_pymodule(mod)
        found_pyname = name_finder.get_pyname_at(code.index('mod') + 1)
        self.assertEqual(mod_pyobject, found_pyname.get_object())

    def test_renaming_functions_with_from_import_and_parens(self):
        mod1 = testutils.create_module(self.project, 'mod1')
        mod1.write('def afunc():\n    pass\n')
        code = 'from mod1 import (\n    afunc as func)\n'
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        mod_pyobject = self.project.get_pymodule(mod1)
        afunc = mod_pyobject['afunc']
        found_pyname = name_finder.get_pyname_at(code.index('afunc') + 1)
        self.assertEqual(afunc.get_object(), found_pyname.get_object())

    @testutils.only_for('2.5')
    def test_relative_modules_after_from_statements(self):
        pkg1 = testutils.create_package(self.project, 'pkg1')
        pkg2 = testutils.create_package(self.project, 'pkg2', pkg1)
        mod1 = testutils.create_module(self.project, 'mod1', pkg1)
        mod2 = testutils.create_module(self.project, 'mod2', pkg2)
        mod1.write('def a_func():\n    pass\n')
        code = 'from ..mod1 import a_func\n'
        mod2.write(code)
        mod2_scope = self.project.get_pymodule(mod2).get_scope()
        name_finder = rope.base.evaluate.ScopeNameFinder(mod2_scope.pyobject)
        mod1_pyobject = self.project.get_pymodule(mod1)
        found_pyname = name_finder.get_pyname_at(code.index('mod1') + 1)
        self.assertEqual(mod1_pyobject, found_pyname.get_object())

    def test_relative_modules_after_from_statements2(self):
        mod1 = testutils.create_module(self.project, 'mod1')
        pkg1 = testutils.create_package(self.project, 'pkg1')
        pkg2 = testutils.create_package(self.project, 'pkg2', pkg1)
        mod2 = testutils.create_module(self.project, 'mod2', pkg2)  # noqa
        mod1.write('import pkg1.pkg2.mod2')

        mod1_scope = self.project.get_pymodule(mod1).get_scope()
        name_finder = rope.base.evaluate.ScopeNameFinder(mod1_scope.pyobject)
        pkg2_pyobject = self.project.get_pymodule(pkg2)
        found_pyname = name_finder.get_pyname_at(mod1.read().index('pkg2') + 1)
        self.assertEqual(pkg2_pyobject, found_pyname.get_object())

    def test_get_pyname_at_on_language_keywords(self):
        code = 'def a_func(a_func):\n    pass\n'
        pymod = libutils.get_string_module(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(pymod)
        with self.assertRaises(exceptions.RopeError):
            name_finder.get_pyname_at(code.index('pass'))

    def test_one_liners(self):
        code = 'var = 1\ndef f(): var = 2\nprint(var)\n'
        pymod = libutils.get_string_module(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(pymod)
        pyname = name_finder.get_pyname_at(code.rindex('var'))
        self.assertEqual(pymod['var'], pyname)

    def test_one_liners_with_line_breaks(self):
        code = 'var = 1\ndef f(\n): var = 2\nprint(var)\n'
        pymod = libutils.get_string_module(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(pymod)
        pyname = name_finder.get_pyname_at(code.rindex('var'))
        self.assertEqual(pymod['var'], pyname)

    def test_one_liners_with_line_breaks2(self):
        code = 'var = 1\ndef f(\np): var = 2\nprint(var)\n'
        pymod = libutils.get_string_module(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(pymod)
        pyname = name_finder.get_pyname_at(code.rindex('var'))
        self.assertEqual(pymod['var'], pyname)


class LogicalLineFinderTest(unittest.TestCase):

    def setUp(self):
        super(LogicalLineFinderTest, self).setUp()

    def tearDown(self):
        super(LogicalLineFinderTest, self).tearDown()

    def _logical_finder(self, code):
        return LogicalLineFinder(SourceLinesAdapter(code))

    def test_normal_lines(self):
        code = 'a_var = 10'
        line_finder = self._logical_finder(code)
        self.assertEqual((1, 1), line_finder.logical_line_in(1))

    def test_normal_lines2(self):
        code = 'another = 10\na_var = 20\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((1, 1), line_finder.logical_line_in(1))
        self.assertEqual((2, 2), line_finder.logical_line_in(2))

    def test_implicit_continuation(self):
        code = 'a_var = 3 + \\\n    4 + \\\n    5'
        line_finder = self._logical_finder(code)
        self.assertEqual((1, 3), line_finder.logical_line_in(2))

    def test_explicit_continuation(self):
        code = 'print(2)\na_var = (3 + \n    4, \n    5)\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((2, 4), line_finder.logical_line_in(2))

    def test_explicit_continuation_comments(self):
        code = '#\na_var = 3\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((2, 2), line_finder.logical_line_in(2))

    def test_multiple_indented_ifs(self):
        code = 'if True:\n    if True:\n        ' \
            'if True:\n            pass\n    a = 10\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((5, 5), line_finder.logical_line_in(5))

    def test_list_comprehensions_and_fors(self):
        code = 'a_list = [i\n    for i in range(10)]\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((1, 2), line_finder.logical_line_in(2))

    def test_generator_expressions_and_fors(self):
        code = 'a_list = (i\n    for i in range(10))\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((1, 2), line_finder.logical_line_in(2))

    def test_fors_and_block_start(self):
        code = 'l = range(10)\nfor i in l:\n    print(i)\n'
        self.assertEqual(2, get_block_start(SourceLinesAdapter(code), 2))

    def test_problems_with_inner_indentations(self):
        code = 'if True:\n    if True:\n        if True:\n            pass\n' \
               '    a = \\\n        1\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((5, 6), line_finder.logical_line_in(6))

    def test_problems_with_inner_indentations2(self):
        code = 'if True:\n    if True:\n        pass\n' \
               'a = 1\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((4, 4), line_finder.logical_line_in(4))

    def test_logical_lines_for_else(self):
        code = 'if True:\n    pass\nelse:\n    pass\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((3, 3), line_finder.logical_line_in(3))

    def test_logical_lines_for_lines_with_wrong_continues(self):
        code = 'var = 1 + \\'
        line_finder = self._logical_finder(code)
        self.assertEqual((1, 1), line_finder.logical_line_in(1))

    def test_logical_lines_for_multiline_string_with_extra_quotes_front(self):
        code = '""""Docs."""\na = 1\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((2, 2), line_finder.logical_line_in(2))

    def test_logical_lines_for_multiline_string_with_escaped_quotes(self):
        code = '"""Quotes \\""" "\\"" \' """\na = 1\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((2, 2), line_finder.logical_line_in(2))

    def test_generating_line_starts(self):
        code = 'a = 1\na = 2\n\na = 3\n'
        line_finder = self._logical_finder(code)
        self.assertEqual([1, 2, 4], list(line_finder.generate_starts()))

    def test_generating_line_starts2(self):
        code = 'a = 1\na = 2\n\na = \\ 3\n'
        line_finder = self._logical_finder(code)
        self.assertEqual([2, 4], list(line_finder.generate_starts(2)))

    def test_generating_line_starts3(self):
        code = 'a = 1\na = 2\n\na = \\ 3\n'
        line_finder = self._logical_finder(code)
        self.assertEqual([2], list(line_finder.generate_starts(2, 3)))

    def test_generating_line_starts_for_multi_line_statements(self):
        code = '\na = \\\n 1 + \\\n 1\n'
        line_finder = self._logical_finder(code)
        self.assertEqual([2], list(line_finder.generate_starts()))

    def test_generating_line_starts_and_unmatched_deindents(self):
        code = 'if True:\n    if True:\n        if True:\n' \
               '            a = 1\n    b = 1\n'
        line_finder = self._logical_finder(code)
        self.assertEqual([4, 5], list(line_finder.generate_starts(4)))

    def test_false_triple_quoted_string(self):
        code = dedent("""\
            def foo():
                a = 0
                p = 'foo'''

            def bar():
                a = 1
                a += 1
        """)
        line_finder = self._logical_finder(code)
        self.assertEqual([1, 2, 3, 5, 6, 7], list(line_finder.generate_starts()))
        self.assertEqual((3, 3), line_finder.logical_line_in(3))
        self.assertEqual([5, 6, 7], list(line_finder.generate_starts(4)))


class TokenizerLogicalLineFinderTest(LogicalLineFinderTest):

    def _logical_finder(self, code):
        lines = SourceLinesAdapter(code)
        return codeanalyze.CachingLogicalLineFinder(
            lines, codeanalyze.tokenizer_generator)


class CustomLogicalLineFinderTest(LogicalLineFinderTest):

    def _logical_finder(self, code):
        lines = SourceLinesAdapter(code)
        return codeanalyze.CachingLogicalLineFinder(
            lines, codeanalyze.custom_generator)


def suite():
    result = unittest.TestSuite()
    result.addTests(unittest.makeSuite(SourceLinesAdapterTest))
    result.addTests(unittest.makeSuite(WordRangeFinderTest))
    result.addTests(unittest.makeSuite(ScopeNameFinderTest))
    result.addTests(unittest.makeSuite(LogicalLineFinderTest))
    result.addTests(unittest.makeSuite(TokenizerLogicalLineFinderTest))
    result.addTests(unittest.makeSuite(CustomLogicalLineFinderTest))
    return result

if __name__ == '__main__':
    unittest.main()



try:
    import unittest2 as unittest
except ImportError:
    import unittest

from rope.base import ast
from rope.refactor import suites


class SuiteTest(unittest.TestCase):

    def setUp(self):
        super(SuiteTest, self).setUp()

    def tearDown(self):
        super(SuiteTest, self).tearDown()

    def test_trivial_case(self):
        root = source_suite_tree('')
        self.assertEqual(1, root.get_start())
        self.assertEqual(0, len(root.get_children()))

    def test_simple_ifs(self):
        root = source_suite_tree('if True:\n    pass')
        self.assertEqual(1, len(root.get_children()))

    def test_simple_else(self):
        root = source_suite_tree(
            'if True:\n    pass\nelse:\n    pass\n')
        self.assertEqual(2, len(root.get_children()))
        self.assertEqual(1, root.get_children()[1].get_start())

    def test_for(self):
        root = source_suite_tree(
            '\nfor i in range(10):\n    pass\nelse:\n    pass\n')
        self.assertEqual(2, len(root.get_children()))
        self.assertEqual(2, root.get_children()[1].get_start())

    def test_while(self):
        root = source_suite_tree(
            'while True:\n    pass\n')
        self.assertEqual(1, len(root.get_children()))
        self.assertEqual(1, root.get_children()[0].get_start())

    def test_with(self):
        root = source_suite_tree(
            'from __future__ import with_statement\nwith file(x):    pass\n')
        self.assertEqual(1, len(root.get_children()))
        self.assertEqual(2, root.get_children()[0].get_start())

    def test_try_finally(self):
        root = source_suite_tree(
            'try:\n    pass\nfinally:\n    pass\n')
        self.assertEqual(2, len(root.get_children()))
        self.assertEqual(1, root.get_children()[0].get_start())

    def test_try_except(self):
        root = source_suite_tree(
            'try:\n    pass\nexcept:\n    pass\nelse:\n    pass\n')
        self.assertEqual(3, len(root.get_children()))
        self.assertEqual(1, root.get_children()[2].get_start())

    def test_try_except_finally(self):
        root = source_suite_tree(
            'try:\n    pass\nexcept:\n    pass\nfinally:\n    pass\n')
        self.assertEqual(3, len(root.get_children()))
        self.assertEqual(1, root.get_children()[2].get_start())

    def test_local_start_and_end(self):
        root = source_suite_tree('if True:\n    pass\nelse:\n    pass\n')
        self.assertEqual(1, root.local_start())
        self.assertEqual(4, root.local_end())
        if_suite = root.get_children()[0]
        self.assertEqual(2, if_suite.local_start())
        self.assertEqual(2, if_suite.local_end())
        else_suite = root.get_children()[1]
        self.assertEqual(4, else_suite.local_start())
        self.assertEqual(4, else_suite.local_end())

    def test_find_suite(self):
        root = source_suite_tree('\n')
        self.assertEqual(root, root.find_suite(1))

    def test_find_suite_for_ifs(self):
        root = source_suite_tree('if True:\n    pass\n')
        if_suite = root.get_children()[0]
        self.assertEqual(if_suite, root.find_suite(2))

    def test_find_suite_for_between_suites(self):
        root = source_suite_tree(
            'if True:\n    pass\nprint(1)\nif True:\n    pass\n')
        if_suite1 = root.get_children()[0]
        if_suite2 = root.get_children()[1]
        self.assertEqual(if_suite1, root.find_suite(2))
        self.assertEqual(if_suite2, root.find_suite(5))
        self.assertEqual(root, root.find_suite(3))

    def test_simple_find_visible(self):
        root = source_suite_tree('a = 1\n')
        self.assertEqual(1, suites.find_visible_for_suite(root, [1]))

    def test_simple_find_visible_ifs(self):
        root = source_suite_tree('\nif True:\n    a = 1\n    b = 2\n')
        self.assertEqual(root.find_suite(3), root.find_suite(4))
        self.assertEqual(3, suites.find_visible_for_suite(root, [3, 4]))

    def test_simple_find_visible_for_else(self):
        root = source_suite_tree('\nif True:\n    pass\nelse:    pass\n')
        self.assertEqual(2, suites.find_visible_for_suite(root, [2, 4]))

    def test_simple_find_visible_for_different_suites(self):
        root = source_suite_tree('if True:\n    pass\na = 1\n'
                                 'if False:\n    pass\n')
        self.assertEqual(1, suites.find_visible_for_suite(root, [2, 3]))
        self.assertEqual(5, suites.find_visible_for_suite(root, [5]))
        self.assertEqual(1, suites.find_visible_for_suite(root, [2, 5]))

    def test_not_always_selecting_scope_start(self):
        root = source_suite_tree(
            'if True:\n    a = 1\n    if True:\n        pass\n'
            '    else:\n        pass\n')
        self.assertEqual(3, suites.find_visible_for_suite(root, [4, 6]))
        self.assertEqual(3, suites.find_visible_for_suite(root, [3, 5]))
        self.assertEqual(3, suites.find_visible_for_suite(root, [4, 5]))

    def test_ignoring_functions(self):
        root = source_suite_tree(
            'def f():\n    pass\na = 1\n')
        self.assertEqual(3, suites.find_visible_for_suite(root, [2, 3]))

    def test_ignoring_classes(self):
        root = source_suite_tree(
            'a = 1\nclass C():\n    pass\n')
        self.assertEqual(1, suites.find_visible_for_suite(root, [1, 3]))


def source_suite_tree(source):
    return suites.ast_suite_tree(ast.parse(source))


if __name__ == '__main__':
    unittest.main()



try:
    import unittest2 as unittest
except ImportError:
    import unittest

import rope.base.history
from rope.base import exceptions
import rope.base.change
from ropetest import testutils


class HistoryTest(unittest.TestCase):

    def setUp(self):
        super(HistoryTest, self).setUp()
        self.project = testutils.sample_project()
        self.history = self.project.history

    def tearDown(self):
        testutils.remove_project(self.project)
        super(HistoryTest, self).tearDown()

    def test_undoing_writes(self):
        my_file = self.project.root.create_file('my_file.txt')
        my_file.write('text1')
        self.history.undo()
        self.assertEqual('', my_file.read())

    def test_moving_files(self):
        my_file = self.project.root.create_file('my_file.txt')
        my_file.move('new_file.txt')
        self.history.undo()
        self.assertEqual('', my_file.read())

    def test_moving_files_to_folders(self):
        my_file = self.project.root.create_file('my_file.txt')
        my_folder = self.project.root.create_folder('my_folder')
        my_file.move(my_folder.path)
        self.history.undo()
        self.assertEqual('', my_file.read())

    def test_writing_files_that_does_not_change_contents(self):
        my_file = self.project.root.create_file('my_file.txt')
        my_file.write('')
        self.project.history.undo()
        self.assertFalse(my_file.exists())


class IsolatedHistoryTest(unittest.TestCase):

    def setUp(self):
        super(IsolatedHistoryTest, self).setUp()
        self.project = testutils.sample_project()
        self.history = rope.base.history.History(self.project)
        self.file1 = self.project.root.create_file('file1.txt')
        self.file2 = self.project.root.create_file('file2.txt')

    def tearDown(self):
        testutils.remove_project(self.project)
        super(IsolatedHistoryTest, self).tearDown()

    def test_simple_undo(self):
        change = rope.base.change.ChangeContents(self.file1, '1')
        self.history.do(change)
        self.assertEqual('1', self.file1.read())
        self.history.undo()
        self.assertEqual('', self.file1.read())

    def test_tobe_undone(self):
        change1 = rope.base.change.ChangeContents(self.file1, '1')
        self.assertEqual(None, self.history.tobe_undone)
        self.history.do(change1)
        self.assertEqual(change1, self.history.tobe_undone)
        change2 = rope.base.change.ChangeContents(self.file1, '2')
        self.history.do(change2)
        self.assertEqual(change2, self.history.tobe_undone)
        self.history.undo()
        self.assertEqual(change1, self.history.tobe_undone)

    def test_tobe_redone(self):
        change = rope.base.change.ChangeContents(self.file1, '1')
        self.history.do(change)
        self.assertEqual(None, self.history.tobe_redone)
        self.history.undo()
        self.assertEqual(change, self.history.tobe_redone)

    def test_undo_limit(self):
        history = rope.base.history.History(self.project, maxundos=1)
        history.do(rope.base.change.ChangeContents(self.file1, '1'))
        history.do(rope.base.change.ChangeContents(self.file1, '2'))
        try:
            history.undo()
            with self.assertRaises(exceptions.HistoryError):
                history.undo()
        finally:
            self.assertEqual('1', self.file1.read())

    def test_simple_redo(self):
        change = rope.base.change.ChangeContents(self.file1, '1')
        self.history.do(change)
        self.history.undo()
        self.history.redo()
        self.assertEqual('1', self.file1.read())

    def test_simple_re_undo(self):
        change = rope.base.change.ChangeContents(self.file1, '1')
        self.history.do(change)
        self.history.undo()
        self.history.redo()
        self.history.undo()
        self.assertEqual('', self.file1.read())

    def test_multiple_undos(self):
        change = rope.base.change.ChangeContents(self.file1, '1')
        self.history.do(change)
        change = rope.base.change.ChangeContents(self.file1, '2')
        self.history.do(change)
        self.history.undo()
        self.assertEqual('1', self.file1.read())
        change = rope.base.change.ChangeContents(self.file1, '3')
        self.history.do(change)
        self.history.undo()
        self.assertEqual('1', self.file1.read())
        self.history.redo()
        self.assertEqual('3', self.file1.read())

    def test_undo_list_underflow(self):
        with self.assertRaises(exceptions.HistoryError):
            self.history.undo()

    def test_redo_list_underflow(self):
        with self.assertRaises(exceptions.HistoryError):
            self.history.redo()

    def test_dropping_undone_changes(self):
        self.file1.write('1')
        with self.assertRaises(exceptions.HistoryError):
            self.history.undo(drop=True)
            self.history.redo()

    def test_undoing_choosen_changes(self):
        change = rope.base.change.ChangeContents(self.file1, '1')
        self.history.do(change)
        self.history.undo(change)
        self.assertEqual('', self.file1.read())
        self.assertFalse(self.history.undo_list)

    def test_undoing_choosen_changes2(self):
        change1 = rope.base.change.ChangeContents(self.file1, '1')
        self.history.do(change1)
        self.history.do(rope.base.change.ChangeContents(self.file1, '2'))
        self.history.undo(change1)
        self.assertEqual('', self.file1.read())
        self.assertFalse(self.history.undo_list)

    def test_undoing_choosen_changes_not_undoing_others(self):
        change1 = rope.base.change.ChangeContents(self.file1, '1')
        self.history.do(change1)
        self.history.do(rope.base.change.ChangeContents(self.file2, '2'))
        self.history.undo(change1)
        self.assertEqual('', self.file1.read())
        self.assertEqual('2', self.file2.read())

    def test_undoing_writing_after_moving(self):
        change1 = rope.base.change.ChangeContents(self.file1, '1')
        self.history.do(change1)
        self.history.do(rope.base.change.MoveResource(self.file1, 'file3.txt'))
        file3 = self.project.get_resource('file3.txt')
        self.history.undo(change1)
        self.assertEqual('', self.file1.read())
        self.assertFalse(file3.exists())

    def test_undoing_folder_movements_for_undoing_writes_inside_it(self):
        folder = self.project.root.create_folder('folder')
        file3 = folder.create_file('file3.txt')
        change1 = rope.base.change.ChangeContents(file3, '1')
        self.history.do(change1)
        self.history.do(rope.base.change.MoveResource(folder, 'new_folder'))
        new_folder = self.project.get_resource('new_folder')
        self.history.undo(change1)
        self.assertEqual('', file3.read())
        self.assertFalse(new_folder.exists())

    def test_undoing_changes_that_depend_on_a_dependant_change(self):
        change1 = rope.base.change.ChangeContents(self.file1, '1')
        self.history.do(change1)
        changes = rope.base.change.ChangeSet('2nd change')
        changes.add_change(rope.base.change.ChangeContents(self.file1, '2'))
        changes.add_change(rope.base.change.ChangeContents(self.file2, '2'))
        self.history.do(changes)
        self.history.do(rope.base.change.MoveResource(self.file2, 'file3.txt'))
        file3 = self.project.get_resource('file3.txt')

        self.history.undo(change1)
        self.assertEqual('', self.file1.read())
        self.assertEqual('', self.file2.read())
        self.assertFalse(file3.exists())

    def test_undoing_writes_for_undoing_folder_movements_containing_it(self):
        folder = self.project.root.create_folder('folder')
        old_file = folder.create_file('file3.txt')
        change1 = rope.base.change.MoveResource(folder, 'new_folder')
        self.history.do(change1)
        new_file = self.project.get_resource('new_folder/file3.txt')
        self.history.do(rope.base.change.ChangeContents(new_file, '1'))
        self.history.undo(change1)
        self.assertEqual('', old_file.read())
        self.assertFalse(new_file.exists())

    def test_undoing_not_available_change(self):
        change = rope.base.change.ChangeContents(self.file1, '1')
        with self.assertRaises(exceptions.HistoryError):
            self.history.undo(change)

    def test_ignoring_ignored_resources(self):
        self.project.set('ignored_resources', ['ignored*'])
        ignored = self.project.get_file('ignored.txt')
        change = rope.base.change.CreateResource(ignored)
        self.history.do(change)
        self.assertTrue(ignored.exists())
        self.assertEqual(0, len(self.history.undo_list))

    def test_get_file_undo_list_simple(self):
        change = rope.base.change.ChangeContents(self.file1, '1')
        self.history.do(change)
        self.assertEqual(set([change]),
                          set(self.history.get_file_undo_list(self.file1)))

    def test_get_file_undo_list_for_moves(self):
        change = rope.base.change.MoveResource(self.file1, 'file2.txt')
        self.history.do(change)
        self.assertEqual(set([change]),
                          set(self.history.get_file_undo_list(self.file1)))

    # XXX: What happens for moves before the file is created?
    def xxx_test_get_file_undo_list_and_moving_its_contining_folder(self):
        folder = self.project.root.create_folder('folder')
        old_file = folder.create_file('file3.txt')
        change1 = rope.base.change.MoveResource(folder, 'new_folder')
        self.history.do(change1)
        self.assertEqual(set([change1]),
                          set(self.history.get_file_undo_list(old_file)))

    def test_clearing_redo_list_after_do(self):
        change = rope.base.change.ChangeContents(self.file1, '1')
        self.history.do(change)
        self.history.undo()
        self.history.do(change)
        self.assertEqual(0, len(self.history.redo_list))

    def test_undoing_a_not_yet_performed_change(self):
        change = rope.base.change.ChangeContents(self.file1, '1')
        str(change)
        with self.assertRaises(exceptions.HistoryError):
            change.undo()

    def test_clearing_up_the_history(self):
        change1 = rope.base.change.ChangeContents(self.file1, '1')
        change2 = rope.base.change.ChangeContents(self.file1, '2')
        self.history.do(change1)
        self.history.do(change2)
        self.history.undo()
        self.history.clear()
        self.assertEqual(0, len(self.history.undo_list))
        self.assertEqual(0, len(self.history.redo_list))

    def test_redoing_choosen_changes_not_undoing_others(self):
        change1 = rope.base.change.ChangeContents(self.file1, '1')
        change2 = rope.base.change.ChangeContents(self.file2, '2')
        self.history.do(change1)
        self.history.do(change2)
        self.history.undo()
        self.history.undo()
        redone = self.history.redo(change2)
        self.assertEqual([change2], redone)
        self.assertEqual('', self.file1.read())
        self.assertEqual('2', self.file2.read())


class SavingHistoryTest(unittest.TestCase):

    def setUp(self):
        super(SavingHistoryTest, self).setUp()
        self.project = testutils.sample_project()
        self.history = rope.base.history.History(self.project)
        self.to_data = rope.base.change.ChangeToData()
        self.to_change = rope.base.change.DataToChange(self.project)

    def tearDown(self):
        testutils.remove_project(self.project)
        super(SavingHistoryTest, self).tearDown()

    def test_simple_set_saving(self):
        data = self.to_data(rope.base.change.ChangeSet('testing'))
        change = self.to_change(data)
        self.assertEqual('testing', str(change))

    def test_simple_change_content_saving(self):
        myfile = self.project.get_file('myfile.txt')
        myfile.create()
        myfile.write('1')
        data = self.to_data(rope.base.change.ChangeContents(myfile, '2'))
        change = self.to_change(data)
        self.history.do(change)
        self.assertEqual('2', myfile.read())
        self.history.undo()
        self.assertEqual('1', change.old_contents)

    def test_move_resource_saving(self):
        myfile = self.project.root.create_file('myfile.txt')
        myfolder = self.project.root.create_folder('myfolder')
        data = self.to_data(rope.base.change.MoveResource(myfile, 'myfolder'))
        change = self.to_change(data)
        self.history.do(change)
        self.assertFalse(myfile.exists())
        self.assertTrue(myfolder.has_child('myfile.txt'))
        self.history.undo()
        self.assertTrue(myfile.exists())
        self.assertFalse(myfolder.has_child('myfile.txt'))

    def test_move_resource_saving_for_folders(self):
        myfolder = self.project.root.create_folder('myfolder')
        newfolder = self.project.get_folder('newfolder')
        change = rope.base.change.MoveResource(myfolder, 'newfolder')
        self.history.do(change)

        data = self.to_data(change)
        change = self.to_change(data)
        change.undo()
        self.assertTrue(myfolder.exists())
        self.assertFalse(newfolder.exists())

    def test_create_file_saving(self):
        myfile = self.project.get_file('myfile.txt')
        data = self.to_data(rope.base.change.CreateFile(self.project.root,
                                                        'myfile.txt'))
        change = self.to_change(data)
        self.history.do(change)
        self.assertTrue(myfile.exists())
        self.history.undo()
        self.assertFalse(myfile.exists())

    def test_create_folder_saving(self):
        myfolder = self.project.get_folder('myfolder')
        data = self.to_data(rope.base.change.CreateFolder(self.project.root,
                                                          'myfolder'))
        change = self.to_change(data)
        self.history.do(change)
        self.assertTrue(myfolder.exists())
        self.history.undo()
        self.assertFalse(myfolder.exists())

    def test_create_resource_saving(self):
        myfile = self.project.get_file('myfile.txt')
        data = self.to_data(rope.base.change.CreateResource(myfile))
        change = self.to_change(data)
        self.history.do(change)
        self.assertTrue(myfile.exists())
        self.history.undo()
        self.assertFalse(myfile.exists())

    def test_remove_resource_saving(self):
        myfile = self.project.root.create_file('myfile.txt')
        data = self.to_data(rope.base.change.RemoveResource(myfile))
        change = self.to_change(data)
        self.history.do(change)
        self.assertFalse(myfile.exists())

    def test_change_set_saving(self):
        change = rope.base.change.ChangeSet('testing')
        myfile = self.project.get_file('myfile.txt')
        change.add_change(rope.base.change.CreateResource(myfile))
        change.add_change(rope.base.change.ChangeContents(myfile, '1'))

        data = self.to_data(change)
        change = self.to_change(data)
        self.history.do(change)
        self.assertEqual('1', myfile.read())
        self.history.undo()
        self.assertFalse(myfile.exists())

    def test_writing_and_reading_history(self):
        history_file = self.project.get_file('history.pickle')  # noqa
        self.project.set('save_history', True)
        history = rope.base.history.History(self.project)
        myfile = self.project.get_file('myfile.txt')
        history.do(rope.base.change.CreateResource(myfile))
        history.write()

        history = rope.base.history.History(self.project)
        history.undo()
        self.assertFalse(myfile.exists())

    def test_writing_and_reading_history2(self):
        history_file = self.project.get_file('history.pickle')  # noqa
        self.project.set('save_history', True)
        history = rope.base.history.History(self.project)
        myfile = self.project.get_file('myfile.txt')
        history.do(rope.base.change.CreateResource(myfile))
        history.undo()
        history.write()

        history = rope.base.history.History(self.project)
        history.redo()
        self.assertTrue(myfile.exists())


def suite():
    result = unittest.TestSuite()
    result.addTests(unittest.makeSuite(HistoryTest))
    result.addTests(unittest.makeSuite(IsolatedHistoryTest))
    result.addTests(unittest.makeSuite(SavingHistoryTest))
    return result

if __name__ == '__main__':
    unittest.main()



"""rope, a python refactoring library"""

INFO = __doc__
VERSION = '0.20.0'
COPYRIGHT = """\
Copyright (C) 2019-2021 Matej Cepl
Copyright (C) 2015-2018 Nicholas Smith
Copyright (C) 2014-2015 Matej Cepl
Copyright (C) 2006-2012 Ali Gholami Rudi
Copyright (C) 2009-2012 Anton Gritsay

This program is free software: you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this program.  If not, see
<https://www.gnu.org/licenses/>."""



"""A module for inferring objects

For more information see the documentation in `rope.base.oi`
package.

"""
import rope.base.builtins
import rope.base.pynames
import rope.base.pyobjects
from rope.base import evaluate, utils, arguments
from rope.base.oi.type_hinting.factory import get_type_hinting_factory


_ignore_inferred = utils.ignore_exception(
    rope.base.pyobjects.IsBeingInferredError)


@_ignore_inferred
def infer_returned_object(pyfunction, args):
    """Infer the `PyObject` this `PyFunction` returns after calling"""
    object_info = pyfunction.pycore.object_info
    result = object_info.get_exact_returned(pyfunction, args)
    if result is not None:
        return result
    result = _infer_returned(pyfunction, args)
    if result is not None:
        if args and pyfunction.get_module().get_resource() is not None:
            params = args.get_arguments(
                pyfunction.get_param_names(special_args=False))
            object_info.function_called(pyfunction, params, result)
        return result
    result = object_info.get_returned(pyfunction, args)
    if result is not None:
        return result
    hint_return = get_type_hinting_factory(pyfunction.pycore.project).make_return_provider()
    type_ = hint_return(pyfunction)
    if type_ is not None:
        return rope.base.pyobjects.PyObject(type_)


@_ignore_inferred
def infer_parameter_objects(pyfunction):
    """Infer the `PyObject` of parameters of this `PyFunction`"""
    object_info = pyfunction.pycore.object_info
    result = object_info.get_parameter_objects(pyfunction)
    if result is None:
        result = _parameter_objects(pyfunction)
    _handle_first_parameter(pyfunction, result)
    return result


def _handle_first_parameter(pyobject, parameters):
    kind = pyobject.get_kind()
    if parameters is None or kind not in ['method', 'classmethod']:
        pass
    if not parameters:
        if not pyobject.get_param_names(special_args=False):
            return
        parameters.append(rope.base.pyobjects.get_unknown())
    if kind == 'method':
        parameters[0] = rope.base.pyobjects.PyObject(pyobject.parent)
    if kind == 'classmethod':
        parameters[0] = pyobject.parent


@_ignore_inferred
def infer_assigned_object(pyname):
    if not pyname.assignments:
        return
    for assignment in reversed(pyname.assignments):
        result = _infer_assignment(assignment, pyname.module)
        if isinstance(result, rope.base.builtins.BuiltinUnknown) and result.get_name() == 'NotImplementedType':
            break
        elif result == rope.base.pyobjects.get_unknown():
            break
        elif result is not None:
            return result

    hint_assignment = get_type_hinting_factory(pyname.module.pycore.project).make_assignment_provider()
    hinting_result = hint_assignment(pyname)
    if hinting_result is not None:
        return rope.base.pyobjects.PyObject(hinting_result)
    return result


def get_passed_objects(pyfunction, parameter_index):
    object_info = pyfunction.pycore.object_info
    result = object_info.get_passed_objects(pyfunction,
                                            parameter_index)
    if not result:
        statically_inferred = _parameter_objects(pyfunction)
        if len(statically_inferred) > parameter_index:
            result.append(statically_inferred[parameter_index])
    return result


def _infer_returned(pyobject, args):
    if args:
        # HACK: Setting parameter objects manually
        # This is not thread safe and might cause problems if `args`
        # does not come from a good call site
        pyobject.get_scope().invalidate_data()
        pyobject._set_parameter_pyobjects(
            args.get_arguments(pyobject.get_param_names(special_args=False)))
    scope = pyobject.get_scope()
    if not scope._get_returned_asts():
        return
    maxtries = 3
    for returned_node in reversed(scope._get_returned_asts()[-maxtries:]):
        try:
            resulting_pyname = evaluate.eval_node(scope, returned_node)
            if resulting_pyname is None:
                continue
            pyobject = resulting_pyname.get_object()
            if pyobject == rope.base.pyobjects.get_unknown():
                continue
            if not scope._is_generator():
                return pyobject
            else:
                return rope.base.builtins.get_generator(pyobject)
        except rope.base.pyobjects.IsBeingInferredError:
            pass


def _parameter_objects(pyobject):
    result = []
    params = pyobject.get_param_names(special_args=False)
    hint_param = get_type_hinting_factory(pyobject.pycore.project).make_param_provider()
    for name in params:
        type_ = hint_param(pyobject, name)
        if type_ is not None:
            result.append(rope.base.pyobjects.PyObject(type_))
        else:
            result.append(rope.base.pyobjects.get_unknown())
    return result

# handling `rope.base.pynames.AssignmentValue`


@_ignore_inferred
def _infer_assignment(assignment, pymodule):
    result = _follow_pyname(assignment, pymodule)
    if result is None:
        return None
    pyname, pyobject = result
    pyobject = _follow_evaluations(assignment, pyname, pyobject)
    if pyobject is None:
        return None
    return _follow_levels(assignment, pyobject)


def _follow_levels(assignment, pyobject):
    for index in assignment.levels:
        if isinstance(pyobject.get_type(), rope.base.builtins.Tuple):
            holdings = pyobject.get_type().get_holding_objects()
            if holdings:
                pyobject = holdings[min(len(holdings) - 1, index)]
            else:
                pyobject = None
        elif isinstance(pyobject.get_type(), rope.base.builtins.List):
            pyobject = pyobject.get_type().holding
        else:
            pyobject = None
        if pyobject is None:
            break
    return pyobject


@_ignore_inferred
def _follow_pyname(assignment, pymodule, lineno=None):
    assign_node = assignment.type_hint or assignment.ast_node
    if lineno is None:
        lineno = _get_lineno_for_node(assign_node)
    holding_scope = pymodule.get_scope().get_inner_scope_for_line(lineno)
    pyname = evaluate.eval_node(holding_scope, assign_node)
    if pyname is not None:
        result = pyname.get_object()
        if isinstance(result.get_type(), rope.base.builtins.Property) and \
           holding_scope.get_kind() == 'Class':
            arg = rope.base.pynames.UnboundName(
                rope.base.pyobjects.PyObject(holding_scope.pyobject))
            return pyname, result.get_type().get_property_object(
                arguments.ObjectArguments([arg]))
        return pyname, result


@_ignore_inferred
def _follow_evaluations(assignment, pyname, pyobject):
    new_pyname = pyname
    tokens = assignment.evaluation.split('.')
    for token in tokens:
        call = token.endswith('()')
        if call:
            token = token[:-2]
        if token:
            pyname = new_pyname
            new_pyname = _get_attribute(pyobject, token)
            if new_pyname is not None:
                pyobject = new_pyname.get_object()
        if pyobject is not None and call:
            if isinstance(pyobject, rope.base.pyobjects.AbstractFunction):
                args = arguments.ObjectArguments([pyname])
                pyobject = pyobject.get_returned_object(args)
            else:
                pyobject = None
        if pyobject is None:
            break
    if pyobject is not None and assignment.assign_type:
        return rope.base.pyobjects.PyObject(pyobject)
    return pyobject


def _get_lineno_for_node(assign_node):
    if hasattr(assign_node, 'lineno') and \
       assign_node.lineno is not None:
        return assign_node.lineno
    return 1


def _get_attribute(pyobject, name):
    if pyobject is not None and name in pyobject:
        return pyobject[name]
